from __future__ import annotations

import pytest

from textual.app import App, ComposeResult
from textual.reactive import var
from textual.widgets import Label, Tab, TabbedContent, TabPane, Tabs
from textual.widgets._tabbed_content import ContentTab


async def test_tabbed_content_switch_via_ui():
    """Check tab navigation via the user interface."""

    class TabbedApp(App):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("foo", id="foo"):
                    yield Label("Foo", id="foo-label")
                with TabPane("bar", id="bar"):
                    yield Label("Bar", id="bar-label")
                with TabPane("baz", id="baz"):
                    yield Label("Baz", id="baz-label")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        # Check first tab
        assert tabbed_content.active == "foo"
        assert tabbed_content.active_pane.id == "foo"
        await pilot.pause()
        assert app.query_one("#foo-label").region
        assert not app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region

        # Click second tab
        await pilot.click(f"Tab#{ContentTab.add_prefix('bar')}")
        assert tabbed_content.active == "bar"
        assert tabbed_content.active_pane.id == "bar"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region

        # Click third tab
        await pilot.click(f"Tab#{ContentTab.add_prefix('baz')}")
        assert tabbed_content.active == "baz"
        assert tabbed_content.active_pane.id == "baz"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert not app.query_one("#bar-label").region
        assert app.query_one("#baz-label").region

        # Press left
        await pilot.press("left")
        assert tabbed_content.active == "bar"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region

        # Press right
        await pilot.press("right")
        assert tabbed_content.active == "baz"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert not app.query_one("#bar-label").region
        assert app.query_one("#baz-label").region


async def test_tabbed_content_switch_via_code():
    """Check tab navigation via code."""

    class TabbedApp(App):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("foo", id="foo"):
                    yield Label("Foo", id="foo-label")
                with TabPane("bar", id="bar"):
                    yield Label("Bar", id="bar-label")
                with TabPane("baz", id="baz"):
                    yield Label("Baz", id="baz-label")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)

        # Check first tab
        assert tabbed_content.active == "foo"
        assert app.query_one("#foo-label").region
        assert not app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region

        # Click second tab
        tabbed_content.active = "bar"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region

        # Click third tab
        tabbed_content.active = "baz"
        await pilot.pause()
        assert not app.query_one("#foo-label").region
        assert not app.query_one("#bar-label").region
        assert app.query_one("#baz-label").region

        # Check fail with non existent tab
        with pytest.raises(ValueError):
            tabbed_content.active = "X"


async def test_unsetting_tabbed_content_active():
    """Check that setting `TabbedContent.active = ""` unsets active tab."""

    messages = []

    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent(initial="bar"):
                with TabPane("foo", id="foo"):
                    yield Label("Foo", id="foo-label")
                with TabPane("bar", id="bar"):
                    yield Label("Bar", id="bar-label")
                with TabPane("baz", id="baz"):
                    yield Label("Baz", id="baz-label")

        def on_tabbed_content_cleared(self, event: TabbedContent.Cleared) -> None:
            messages.append(event)

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        assert bool(tabbed_content.active)
        tabbed_content.active = ""
        await pilot.pause()
        assert len(messages) == 1
        assert isinstance(messages[0], TabbedContent.Cleared)


async def test_tabbed_content_initial():
    """Checked tabbed content with non-default tab."""

    class TabbedApp(App):
        def compose(self) -> ComposeResult:
            with TabbedContent(initial="bar"):
                with TabPane("foo", id="foo"):
                    yield Label("Foo", id="foo-label")
                with TabPane("bar", id="bar"):
                    yield Label("Bar", id="bar-label")
                with TabPane("baz", id="baz"):
                    yield Label("Baz", id="baz-label")

    app = TabbedApp()
    async with app.run_test():
        tabbed_content = app.query_one(TabbedContent)
        assert tabbed_content.active == "bar"

        # Check only bar is visible
        assert not app.query_one("#foo-label").region
        assert app.query_one("#bar-label").region
        assert not app.query_one("#baz-label").region


async def test_tabbed_content_messages():
    class TabbedApp(App):
        activation_history: list[Tab] = []

        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("foo", id="foo"):
                    yield Label("Foo", id="foo-label")
                with TabPane("bar", id="bar"):
                    yield Label("Bar", id="bar-label")
                with TabPane("baz", id="baz"):
                    yield Label("Baz", id="baz-label")

        def on_tabbed_content_tab_activated(
            self, event: TabbedContent.TabActivated
        ) -> None:
            self.activation_history.append(event.tab)

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        tabbed_content.active = "bar"
        await pilot.pause()
        assert app.activation_history == [
            # foo was originally activated.
            app.query_one(TabbedContent).get_tab("foo"),
            # then we did bar "by hand"
            app.query_one(TabbedContent).get_tab("bar"),
        ]


async def test_tabbed_content_add_later_from_empty():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            yield TabbedContent()

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.active == ""
        assert tabbed_content.tab_count == 0
        await tabbed_content.add_pane(TabPane("Test 1", id="test-1"))
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "test-1"
        await tabbed_content.add_pane(TabPane("Test 2", id="test-2"))
        assert tabbed_content.tab_count == 2
        assert tabbed_content.active == "test-1"


async def test_tabbed_content_add_later_from_composed():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")
                yield TabPane("Test 2", id="initial-2")
                yield TabPane("Test 3", id="initial-3")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 3
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(TabPane("Test 4", id="test-1"))
        assert tabbed_content.tab_count == 4
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(TabPane("Test 5", id="test-2"))
        assert tabbed_content.tab_count == 5
        assert tabbed_content.active == "initial-1"


async def test_tabbed_content_add_before_id():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(TabPane("Added", id="new-1"), before="initial-1")
        assert tabbed_content.tab_count == 2
        assert tabbed_content.active == "initial-1"
        assert [
            ContentTab.sans_prefix(tab.id) for tab in tabbed_content.query(Tab)
        ] == [
            "new-1",
            "initial-1",
        ]


async def test_tabbed_content_add_before_pane():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(
            TabPane("Added", id="new-1"),
            before=pilot.app.query_one("TabPane#initial-1", TabPane),
        )
        assert tabbed_content.tab_count == 2
        assert tabbed_content.active == "initial-1"
        assert [
            ContentTab.sans_prefix(tab.id)
            for tab in tabbed_content.query(Tab).results(Tab)
        ] == [
            "new-1",
            "initial-1",
        ]


async def test_tabbed_content_add_before_badly():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        with pytest.raises(Tabs.TabError):
            await tabbed_content.add_pane(
                TabPane("Added", id="new-1"), before="unknown-1"
            )


async def test_tabbed_content_add_after():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(TabPane("Added", id="new-1"), after="initial-1")
        assert tabbed_content.tab_count == 2
        assert tabbed_content.active == "initial-1"
        assert [
            ContentTab.sans_prefix(tab.id)
            for tab in tabbed_content.query(Tab).results(Tab)
        ] == [
            "initial-1",
            "new-1",
        ]


async def test_tabbed_content_add_after_pane():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        await tabbed_content.add_pane(
            TabPane("Added", id="new-1"),
            after=pilot.app.query_one("TabPane#initial-1", TabPane),
        )
        await pilot.pause()
        assert tabbed_content.tab_count == 2
        assert tabbed_content.active == "initial-1"
        assert [
            ContentTab.sans_prefix(tab.id)
            for tab in tabbed_content.query(Tab).results(Tab)
        ] == [
            "initial-1",
            "new-1",
        ]


async def test_tabbed_content_add_after_badly():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        with pytest.raises(Tabs.TabError):
            await tabbed_content.add_pane(
                TabPane("Added", id="new-1"), after="unknown-1"
            )


async def test_tabbed_content_add_before_and_after():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 1
        assert tabbed_content.active == "initial-1"
        with pytest.raises(Tabs.TabError):
            await tabbed_content.add_pane(
                TabPane("Added", id="new-1"), before="initial-1", after="initial-1"
            )


async def test_tabbed_content_removal():
    class TabbedApp(App[None]):
        cleared: var[int] = var(0)

        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")
                yield TabPane("Test 2", id="initial-2")
                yield TabPane("Test 3", id="initial-3")

        def on_tabbed_content_cleared(self) -> None:
            self.cleared += 1

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 3
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-1"
        await tabbed_content.remove_pane("initial-1")
        await pilot.pause()
        assert tabbed_content.tab_count == 2
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-2"
        await tabbed_content.remove_pane("initial-2")
        await pilot.pause()
        assert tabbed_content.tab_count == 1
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-3"
        await tabbed_content.remove_pane("initial-3")
        await pilot.pause()
        assert tabbed_content.tab_count == 0
        assert pilot.app.cleared == 1
        assert tabbed_content.active == ""


async def test_tabbed_content_reversed_removal():
    class TabbedApp(App[None]):
        cleared: var[int] = var(0)

        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")
                yield TabPane("Test 2", id="initial-2")
                yield TabPane("Test 3", id="initial-3")

        def on_tabbed_content_cleared(self) -> None:
            self.cleared += 1

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 3
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-1"
        await tabbed_content.remove_pane("initial-3")
        await pilot.pause()
        assert tabbed_content.tab_count == 2
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-1"
        await tabbed_content.remove_pane("initial-2")
        await pilot.pause()
        assert tabbed_content.tab_count == 1
        assert pilot.app.cleared == 0
        assert tabbed_content.active == "initial-1"
        await tabbed_content.remove_pane("initial-1")
        await pilot.pause()
        assert tabbed_content.tab_count == 0
        assert pilot.app.cleared == 1
        assert tabbed_content.active == ""


async def test_tabbed_content_clear():
    class TabbedApp(App[None]):
        cleared: var[int] = var(0)

        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("Test 1", id="initial-1")
                yield TabPane("Test 2", id="initial-2")
                yield TabPane("Test 3", id="initial-3")

        def on_tabbed_content_cleared(self) -> None:
            self.cleared += 1

    async with TabbedApp().run_test() as pilot:
        tabbed_content = pilot.app.query_one(TabbedContent)
        assert tabbed_content.tab_count == 3
        assert tabbed_content.active == "initial-1"
        assert pilot.app.cleared == 0
        await tabbed_content.clear_panes()
        await pilot.pause()
        assert tabbed_content.tab_count == 0
        assert tabbed_content.active == ""
        assert pilot.app.cleared == 1


async def test_disabling_does_not_deactivate_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).get_tab("tab-1").disabled = True

    app = TabbedApp()
    async with app.run_test():
        assert app.query_one(TabbedContent).active == "tab-1"


async def test_disabled_tab_cannot_be_clicked():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).get_tab("tab-2").disabled = True

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"


async def test_disabling_via_tabbed_content():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).disable_tab("tab-2")

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"


async def test_disabling_via_tab_pane():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one("TabPane#tab-2").disabled = True

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"


async def test_creating_disabled_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("first"):
                    yield Label("hello")
                with TabPane("second", disabled=True):
                    yield Label("world")

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"


async def test_navigation_around_disabled_tabs():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")
                yield Label("tab-3")
                yield Label("tab-4")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).get_tab("tab-1").disabled = True
            self.query_one(TabbedContent).get_tab("tab-3").disabled = True

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_conent = app.query_one(TabbedContent)
        tabs = app.query_one(Tabs)
        assert tabbed_conent.active == "tab-1"
        tabs.action_next_tab()
        await pilot.pause()
        assert tabbed_conent.active == "tab-2"
        tabs.action_next_tab()
        await pilot.pause()
        assert tabbed_conent.active == "tab-4"
        tabs.action_next_tab()
        await pilot.pause()
        assert tabbed_conent.active == "tab-2"
        tabs.action_previous_tab()
        await pilot.pause()
        assert tabbed_conent.active == "tab-4"


async def test_reenabling_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).get_tab("tab-2").disabled = True

        def reenable(self) -> None:
            self.query_one(TabbedContent).get_tab("tab-2").disabled = False

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"
        app.reenable()
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-2"


async def test_reenabling_via_tabbed_content():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).disable_tab("tab-2")

        def reenable(self) -> None:
            self.query_one(TabbedContent).enable_tab("tab-2")

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"
        app.reenable()
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-2"


async def test_reenabling_via_tab_pane():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

        def on_mount(self) -> None:
            self.query_one("TabPane#tab-2").disabled = True

        def reenable(self) -> None:
            self.query_one("TabPane#tab-2").disabled = False

    app = TabbedApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-1"
        app.reenable()
        await pilot.click(f"Tab#{ContentTab.add_prefix('tab-2')}")
        assert app.query_one(TabbedContent).active == "tab-2"


async def test_disabling_unknown_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

    app = TabbedApp()
    async with app.run_test():
        with pytest.raises(Tabs.TabError):
            app.query_one(TabbedContent).disable_tab("foo")


async def test_enabling_unknown_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

    app = TabbedApp()
    async with app.run_test():
        with pytest.raises(Tabs.TabError):
            app.query_one(TabbedContent).enable_tab("foo")


async def test_hide_unknown_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

    app = TabbedApp()
    async with app.run_test():
        with pytest.raises(Tabs.TabError):
            app.query_one(TabbedContent).hide_tab("foo")


async def test_show_unknown_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

    app = TabbedApp()
    async with app.run_test():
        with pytest.raises(Tabs.TabError):
            app.query_one(TabbedContent).show_tab("foo")


async def test_hide_show_messages():
    hide_msg = False
    show_msg = False

    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

        def on_tabs_tab_hidden(self) -> None:
            nonlocal hide_msg
            hide_msg = True

        def on_tabs_tab_shown(self) -> None:
            nonlocal show_msg
            show_msg = True

    app = TabbedApp()
    async with app.run_test() as pilot:
        app.query_one(TabbedContent).hide_tab("tab-1")
        await pilot.pause()
        assert hide_msg
        app.query_one(TabbedContent).show_tab("tab-1")
        await pilot.pause()
        assert show_msg


async def test_hide_last_tab_means_no_tab_active():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        tabbed_content.hide_tab("tab-1")
        await pilot.pause()
        assert not tabbed_content.active


async def test_hiding_tabs_moves_active_to_next_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")
                yield Label("tab-3")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        tabbed_content.hide_tab("tab-1")
        await pilot.pause()
        assert tabbed_content.active == "tab-2"
        tabbed_content.hide_tab("tab-2")
        await pilot.pause()
        assert tabbed_content.active == "tab-3"


async def test_showing_tabs_does_not_change_active_tab():
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")
                yield Label("tab-3")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        tabbed_content.hide_tab("tab-1")
        tabbed_content.hide_tab("tab-2")
        await pilot.pause()
        # sanity check
        assert tabbed_content.active == "tab-3"

        tabbed_content.show_tab("tab-1")
        tabbed_content.show_tab("tab-2")
        assert tabbed_content.active == "tab-3"


@pytest.mark.parametrize("tab_id", ["tab-1", "tab-2"])
async def test_showing_first_tab_activates_tab(tab_id: str):
    class TabbedApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield Label("tab-1")
                yield Label("tab-2")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabbed_content = app.query_one(TabbedContent)
        tabbed_content.hide_tab("tab-1")
        tabbed_content.hide_tab("tab-2")
        await pilot.pause()
        # sanity check
        assert not tabbed_content.active

        tabbed_content.show_tab(tab_id)
        await pilot.pause()
        assert tabbed_content.active == tab_id


async def test_disabling_nested_tabs():
    """Regression test for https://github.com/Textualize/textual/issues/3145."""

    class TabbedApp(App):
        def compose(self) -> ComposeResult:
            with TabbedContent(id="tabbed-content"):
                with TabPane("Tab Pane 1"):
                    yield Label("foo")
                with TabPane("Tab Pane 2"):
                    yield Label("bar")
                with TabPane("Tab Pane 3"):
                    with TabbedContent():
                        with TabPane("Inner Pane 1"):
                            yield Label("fizz")
                        with TabPane("Inner Pane 2"):
                            yield Label("bang")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabber = app.query_one("#tabbed-content", expect_type=TabbedContent)
        tabber.disable_tab("tab-1")
        await pilot.pause()
        tabber.enable_tab("tab-1")
        await pilot.pause()


async def test_hiding_nested_tabs():
    """Regression test for https://github.com/Textualize/textual/issues/3145."""

    class TabbedApp(App):
        def compose(self) -> ComposeResult:
            with TabbedContent(id="tabbed-content"):
                with TabPane("Tab Pane 1"):
                    yield Label("foo")
                with TabPane("Tab Pane 2"):
                    yield Label("bar")
                with TabPane("Tab Pane 3"):
                    with TabbedContent():
                        with TabPane("Inner Pane 1"):
                            yield Label("fizz")
                        with TabPane("Inner Pane 2"):
                            yield Label("bang")

    app = TabbedApp()
    async with app.run_test() as pilot:
        tabber = app.query_one("#tabbed-content", expect_type=TabbedContent)
        tabber.hide_tab("tab-1")
        await pilot.pause()
        tabber.show_tab("tab-1")
        await pilot.pause()


async def test_tabs_nested_in_tabbed_content_doesnt_crash():
    """Regression test for https://github.com/Textualize/textual/issues/3412"""

    class TabsNestedInTabbedContent(App):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("Outer TabPane"):
                    yield Tabs("Inner Tab")

    app = TabsNestedInTabbedContent()
    async with app.run_test() as pilot:
        await pilot.pause()


async def test_tabs_nested_doesnt_interfere_with_ancestor_tabbed_content():
    """When a Tabs is nested as a descendant in the DOM of a TabbedContent,
    the messages posted from that Tabs should not interfere with the TabbedContent.
    A TabbedContent should only handle messages from Tabs which are direct children.

    Relates to https://github.com/Textualize/textual/issues/3412
    """

    class TabsNestedInTabbedContent(App):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("OuterTab", id="outer1"):
                    yield Tabs(
                        Tab("Tab1", id="tab1"),
                        Tab("Tab2", id="tab2"),
                        id="inner-tabs",
                    )

    app = TabsNestedInTabbedContent()
    async with app.run_test():
        inner_tabs = app.query_one("#inner-tabs", Tabs)
        tabbed_content = app.query_one(TabbedContent)

        assert inner_tabs.active_tab.id == "tab1"
        assert tabbed_content.active == "outer1"

        await inner_tabs.clear()

        assert inner_tabs.active_tab is None
        assert tabbed_content.active == "outer1"


async def test_disabling_tab_within_tabbed_content_stays_isolated():
    """Disabling a tab within a tab pane should not affect the TabbedContent."""

    class TabsNestedInTabbedContent(App):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                with TabPane("TabbedContent", id="duplicate"):
                    yield Tabs(
                        Tab("Tab1", id="duplicate"),
                        Tab("Tab2", id="stay-enabled"),
                        id="test-tabs",
                    )

    app = TabsNestedInTabbedContent()
    async with app.run_test() as pilot:
        assert app.query_one("Tab#duplicate").disabled is False
        assert app.query_one("TabPane#duplicate").disabled is False
        app.query_one("#test-tabs", Tabs).disable("duplicate")
        await pilot.pause()
        assert app.query_one("Tab#duplicate").disabled is True
        assert app.query_one("TabPane#duplicate").disabled is False



from textual.app import App, ComposeResult
from textual.widgets import Footer, Label


class ToggleCompactFooterApp(App):
    CSS = """
    Screen {
        align: center middle;
    }
    """

    BINDINGS = [
        ("ctrl+t", "toggle_compact_footer", "Toggle Compact Footer"),
    ]

    def compose(self) -> ComposeResult:
        yield Label("Compact Footer")
        yield Footer()

    def on_mount(self) -> None:
        footer = self.query_one(Footer)
        footer.compact = True

    def action_toggle_compact_footer(self) -> None:
        footer = self.query_one(Footer)
        footer.compact = not footer.compact
        footer_type = "Compact" if footer.compact else "Standard"
        self.query_one(Label).update(f"{footer_type} Footer")


if __name__ == "__main__":
    app = ToggleCompactFooterApp()
    app.run()



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.command import Hit, Hits, Provider
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Input


class SimpleSource(Provider):
    environment: set[tuple[App, Screen, Widget | None]] = set()

    async def search(self, _: str) -> Hits:
        def goes_nowhere_does_nothing() -> None:
            pass

        SimpleSource.environment.add((self.app, self.screen, self.focused))
        yield Hit(1, "Hit", goes_nowhere_does_nothing, "Hit")


class CommandPaletteApp(App[None]):
    COMMANDS = {SimpleSource}

    def compose(self) -> ComposeResult:
        yield Input()

    def on_mount(self) -> None:
        self.action_command_palette()


async def test_command_source_environment() -> None:
    """The command source should see the app and default screen."""
    async with CommandPaletteApp().run_test() as pilot:
        base_screen, *_ = pilot.app.children
        assert base_screen is not None
        await pilot.press(*"test")
        assert len(SimpleSource.environment) == 1
        assert SimpleSource.environment == {
            (pilot.app, base_screen, base_screen.query_one(Input))
        }



"""Regression test for #1616 https://github.com/Textualize/textual/issues/1616"""

from textual.app import App
from textual.containers import VerticalScroll


async def test_overflow_change_updates_virtual_size_appropriately():
    class MyApp(App):
        def compose(self):
            yield VerticalScroll()

    app = MyApp()

    async with app.run_test() as pilot:
        vertical = app.query_one(VerticalScroll)

        height = vertical.virtual_size.height

        vertical.styles.overflow_x = "scroll"
        await pilot.pause()  # Let changes propagate.
        assert vertical.virtual_size.height < height

        vertical.styles.overflow_x = "hidden"
        await pilot.pause()
        assert vertical.virtual_size.height == height

        width = vertical.virtual_size.width

        vertical.styles.overflow_y = "scroll"
        await pilot.pause()
        assert vertical.virtual_size.width < width

        vertical.styles.overflow_y = "hidden"
        await pilot.pause()
        assert vertical.virtual_size.width == width



from textual.app import App, ComposeResult
from textual.widgets import Label, TabbedContent, TabPane


class ColorTabsApp(App):
    CSS = """
    TabbedContent #--content-tab-green {
        color: green;
    }

    TabbedContent #--content-tab-red {
        color: red;
    }
    """

    def compose(self) -> ComposeResult:
        with TabbedContent():
            with TabPane("Red", id="red"):
                yield Label("Red!")
            with TabPane("Green", id="green"):
                yield Label("Green!")


if __name__ == "__main__":
    ColorTabsApp().run()



"""Unit tests aimed at testing the option list messages."""

from __future__ import annotations

from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import OptionList
from textual.widgets.option_list import Option


class OptionListApp(App[None]):
    """Test option list application."""

    def __init__(self) -> None:
        super().__init__()
        self.messages: list[tuple[str, str, int]] = []

    def compose(self) -> ComposeResult:
        yield OptionList(*[Option(str(n), id=str(n)) for n in range(10)])

    def _record(self, event: OptionList.OptionMessage) -> None:
        assert isinstance(event.option_id, str)
        assert event.option_list is event.control
        self.messages.append(
            (event.__class__.__name__, event.option_id, event.option_index)
        )

    def on_option_list_option_highlighted(
        self, event: OptionList.OptionHighlighted
    ) -> None:
        self._record(event)

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        self._record(event)


async def test_messages_on_startup() -> None:
    """There should be a highlighted message when a non-empty option list first starts up."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.pause()
        assert pilot.app.messages == [("OptionHighlighted", "0", 0)]


async def test_same_highlight_message() -> None:
    """Highlighting a highlight should result in no message."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.pause()
        pilot.app.query_one(OptionList).highlighted = 0
        await pilot.pause()
        assert pilot.app.messages == [("OptionHighlighted", "0", 0)]


async def test_highlight_disabled_option_no_message() -> None:
    """Highlighting a disabled option should result in no messages."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.pause()
        pilot.app.query_one(OptionList).disable_option("1")
        pilot.app.query_one(OptionList).highlighted = 1
        await pilot.pause()
        assert pilot.app.messages[1:] == []


async def test_new_highlight() -> None:
    """Setting the highlight to a new option should result in a message."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.pause()
        pilot.app.query_one(OptionList).highlighted = 2
        await pilot.pause()
        assert pilot.app.messages[1:] == [("OptionHighlighted", "2", 2)]


async def test_move_highlight_with_keyboard() -> None:
    """Changing option via the keyboard should result in a message."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.press("tab", "down")
        assert pilot.app.messages[1:] == [("OptionHighlighted", "1", 1)]


async def test_select_message_with_keyboard() -> None:
    """Hitting enter on an option should result in a message."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.press("tab", "down", "enter")
        assert pilot.app.messages[1:] == [
            ("OptionHighlighted", "1", 1),
            ("OptionSelected", "1", 1),
        ]


async def test_click_option_with_mouse() -> None:
    """Clicking on an option via the mouse should result in highlight and select messages."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        await pilot.click(OptionList, Offset(2, 2))
        assert pilot.app.messages[1:] == [
            ("OptionHighlighted", "1", 1),
            ("OptionSelected", "1", 1),
        ]


async def test_click_disabled_option_with_mouse() -> None:
    """Clicking on a disabled option via the mouse should result no messages."""
    async with OptionListApp().run_test() as pilot:
        assert isinstance(pilot.app, OptionListApp)
        pilot.app.query_one(OptionList).disable_option("1")
        await pilot.click(OptionList, Offset(1, 1))
        assert pilot.app.messages[1:] == []



from textual.app import App, ComposeResult
from textual.widgets import Pretty

DATA = {
    "title": "Back to the Future",
    "releaseYear": 1985,
    "director": "Robert Zemeckis",
    "genre": "Adventure, Comedy, Sci-Fi",
    "cast": [
        {"actor": "Michael J. Fox", "character": "Marty McFly"},
        {"actor": "Christopher Lloyd", "character": "Dr. Emmett Brown"},
    ],
}


class PrettyExample(App):
    def compose(self) -> ComposeResult:
        yield Pretty(DATA)


app = PrettyExample()

if __name__ == "__main__":
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Button, Label, TabbedContent, Tabs, TabPane

class TabbedContentStyleLeakTestApp(App[None]):

    def compose(self) -> ComposeResult:
        with TabbedContent():
            with TabPane("Leak Test"):
                yield Label("This label should come first")
                yield Button("This button should come second")
                yield Tabs("These", "Tabs", "Should", "Come", "Last")

if __name__ == "__main__":
    TabbedContentStyleLeakTestApp().run()



"""

This module provides a number of classes for validating input.

See [Validating Input](/widgets/input/#validating-input) for details.

"""

from __future__ import annotations

import math
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Pattern, Sequence
from urllib.parse import urlparse

import rich.repr


@dataclass
class ValidationResult:
    """The result of calling a `Validator.validate` method."""

    failures: Sequence[Failure] = field(default_factory=list)
    """A list of reasons why the value was invalid. Empty if valid=True"""

    @staticmethod
    def merge(results: Sequence["ValidationResult"]) -> "ValidationResult":
        """Merge multiple ValidationResult objects into one.

        Args:
            results: List of ValidationResult objects to merge.

        Returns:
            Merged ValidationResult object.
        """
        is_valid = all(result.is_valid for result in results)
        failures = [failure for result in results for failure in result.failures]
        if is_valid:
            return ValidationResult.success()
        else:
            return ValidationResult.failure(failures)

    @staticmethod
    def success() -> ValidationResult:
        """Construct a successful ValidationResult.

        Returns:
            A successful ValidationResult.
        """
        return ValidationResult()

    @staticmethod
    def failure(failures: Sequence[Failure]) -> ValidationResult:
        """Construct a failure ValidationResult.

        Args:
            failures: The failures.

        Returns:
            A failure ValidationResult.
        """
        return ValidationResult(failures)

    @property
    def failure_descriptions(self) -> list[str]:
        """Utility for extracting failure descriptions as strings.

        Useful if you don't care about the additional metadata included in the `Failure` objects.

        Returns:
            A list of the string descriptions explaining the failing validations.
        """
        return [
            failure.description
            for failure in self.failures
            if failure.description is not None
        ]

    @property
    def is_valid(self) -> bool:
        """True if the validation was successful."""
        return len(self.failures) == 0


@dataclass
class Failure:
    """Information about a validation failure."""

    validator: Validator
    """The Validator which produced the failure."""
    value: str | None = None
    """The value which resulted in validation failing."""
    description: str | None = None
    """An optional override for describing this failure. Takes precedence over any messages set in the Validator."""

    def __post_init__(self) -> None:
        # If a failure message isn't supplied, try to get it from the Validator.
        if self.description is None:
            if self.validator.failure_description is not None:
                self.description = self.validator.failure_description
            else:
                self.description = self.validator.describe_failure(self)

    def __rich_repr__(self) -> rich.repr.Result:  # pragma: no cover
        yield self.value
        yield self.validator
        yield self.description


class Validator(ABC):
    """Base class for the validation of string values.

    Commonly used in conjunction with the `Input` widget, which accepts a
    list of validators via its constructor. This validation framework can also be used to validate any 'stringly-typed'
    values (for example raw command line input from `sys.args`).

    To implement your own `Validator`, subclass this class.

    Example:
        ```python
        class Palindrome(Validator):
            def validate(self, value: str) -> ValidationResult:
                def is_palindrome(value: str) -> bool:
                    return value == value[::-1]
                return self.success() if is_palindrome(value) else self.failure("Not palindrome!")
        ```
    """

    def __init__(self, failure_description: str | None = None) -> None:
        self.failure_description = failure_description
        """A description of why the validation failed.

        The description (intended to be user-facing) to attached to the Failure if the validation fails.
        This failure description is ultimately accessible at the time of validation failure  via the `Input.Changed`
        or `Input.Submitted` event, and you can access it on your message handler (a method called, for example,
        `on_input_changed` or a method decorated with `@on(Input.Changed)`.
        """

    @abstractmethod
    def validate(self, value: str) -> ValidationResult:
        """Validate the value and return a ValidationResult describing the outcome of the validation.

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """

    def describe_failure(self, failure: Failure) -> str | None:
        """Return a string description of the Failure.

        Used to provide a more fine-grained description of the failure. A Validator could fail for multiple
        reasons, so this method could be used to provide a different reason for different types of failure.

        !!! warning

            This method is only called if no other description has been supplied. If you supply a description
            inside a call to `self.failure(description="...")`, or pass a description into the constructor of
            the validator, those will take priority, and this method won't be called.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        return self.failure_description

    def success(self) -> ValidationResult:
        """Shorthand for `ValidationResult(True)`.

        You can return success() from a `Validator.validate` method implementation to signal
        that validation has succeeded.

        Returns:
            A ValidationResult indicating validation succeeded.
        """
        return ValidationResult()

    def failure(
        self,
        description: str | None = None,
        value: str | None = None,
        failures: Failure | Sequence[Failure] | None = None,
    ) -> ValidationResult:
        """Shorthand for signaling validation failure.

        You can return failure(...) from a `Validator.validate` implementation to signal validation succeeded.

        Args:
            description: The failure description that will be used. When used in conjunction with the Input widget,
                this is the description that will ultimately be available inside the handler for `Input.Changed`. If not
                supplied, the `failure_description` from the `Validator` will be used. If that is not supplied either,
                then the `describe_failure` method on `Validator` will be called.
            value: The value that was considered invalid. This is optional, and only needs to be supplied if required
                in your `Input.Changed` handler.
            failures: The reasons the validator failed. If not supplied, a generic `Failure` will be included in the
                ValidationResult returned from this function.

        Returns:
            A ValidationResult representing failed validation, and containing the metadata supplied
                to this function.
        """
        if isinstance(failures, Failure):
            failures = [failures]

        result = ValidationResult(
            failures or [Failure(validator=self, value=value, description=description)],
        )
        return result


class Regex(Validator):
    """A validator that checks the value matches a regex (via `re.fullmatch`)."""

    def __init__(
        self,
        regex: str | Pattern[str],
        flags: int | re.RegexFlag = 0,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)
        self.regex = regex
        """The regex which we'll validate is matched by the value."""
        self.flags = flags
        """The flags to pass to `re.fullmatch`."""

    class NoResults(Failure):
        """Indicates validation failed because the regex could not be found within the value string."""

    def validate(self, value: str) -> ValidationResult:
        """Ensure that the value matches the regex.

        Args:
            value: The value that should match the regex.

        Returns:
            The result of the validation.
        """
        regex = self.regex
        has_match = re.fullmatch(regex, value, flags=self.flags) is not None
        if not has_match:
            failures = [Regex.NoResults(self, value)]
            return self.failure(failures=failures)
        return self.success()

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        return f"Must match regular expression {self.regex!r} (flags={self.flags})."


class Number(Validator):
    """Validator that ensures the value is a number, with an optional range check."""

    def __init__(
        self,
        minimum: float | None = None,
        maximum: float | None = None,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)
        self.minimum = minimum
        """The minimum value of the number, inclusive. If `None`, the minimum is unbounded."""
        self.maximum = maximum
        """The maximum value of the number, inclusive. If `None`, the maximum is unbounded."""

    class NotANumber(Failure):
        """Indicates a failure due to the value not being a valid number (decimal/integer, inc. scientific notation)"""

    class NotInRange(Failure):
        """Indicates a failure due to the number not being within the range [minimum, maximum]."""

    def validate(self, value: str) -> ValidationResult:
        """Ensure that `value` is a valid number, optionally within a range.

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        try:
            float_value = float(value)
        except ValueError:
            return ValidationResult.failure([Number.NotANumber(self, value)])

        if math.isnan(float_value) or math.isinf(float_value):
            return ValidationResult.failure([Number.NotANumber(self, value)])

        if not self._validate_range(float_value):
            return ValidationResult.failure(
                [Number.NotInRange(self, value)],
            )
        return self.success()

    def _validate_range(self, value: float) -> bool:
        """Return a boolean indicating whether the number is within the range specified in the attributes."""
        if self.minimum is not None and value < self.minimum:
            return False
        if self.maximum is not None and value > self.maximum:
            return False
        return True

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        if isinstance(failure, Number.NotANumber):
            return "Must be a valid number."
        elif isinstance(failure, Number.NotInRange):
            if self.minimum is None and self.maximum is not None:
                return f"Must be less than or equal to {self.maximum}."
            elif self.minimum is not None and self.maximum is None:
                return f"Must be greater than or equal to {self.minimum}."
            else:
                return f"Must be between {self.minimum} and {self.maximum}."
        else:
            return None


class Integer(Number):
    """Validator which ensures the value is an integer which falls within a range."""

    class NotAnInteger(Failure):
        """Indicates a failure due to the value not being a valid integer."""

    def validate(self, value: str) -> ValidationResult:
        """Ensure that `value` is an integer, optionally within a range.

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        # First, check that we're dealing with a number in the range.
        number_validation_result = super().validate(value)
        if not number_validation_result.is_valid:
            return number_validation_result

        # We know it's a number, but is that number an integer?
        try:
            int_value = int(value)
        except ValueError:
            return ValidationResult.failure([Integer.NotAnInteger(self, value)])
        return self.success()

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        if isinstance(failure, (Integer.NotANumber, Integer.NotAnInteger)):
            return "Must be a valid integer."
        elif isinstance(failure, Integer.NotInRange):
            if self.minimum is None and self.maximum is not None:
                return f"Must be less than or equal to {self.maximum}."
            elif self.minimum is not None and self.maximum is None:
                return f"Must be greater than or equal to {self.minimum}."
            else:
                return f"Must be between {self.minimum} and {self.maximum}."
        else:
            return None


class Length(Validator):
    """Validate that a string is within a range (inclusive)."""

    def __init__(
        self,
        minimum: int | None = None,
        maximum: int | None = None,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)
        self.minimum = minimum
        """The inclusive minimum length of the value, or None if unbounded."""
        self.maximum = maximum
        """The inclusive maximum length of the value, or None if unbounded."""

    class Incorrect(Failure):
        """Indicates a failure due to the length of the value being outside the range."""

    def validate(self, value: str) -> ValidationResult:
        """Ensure that value falls within the maximum and minimum length constraints.

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        too_short = self.minimum is not None and len(value) < self.minimum
        too_long = self.maximum is not None and len(value) > self.maximum
        if too_short or too_long:
            return ValidationResult.failure([Length.Incorrect(self, value)])
        return self.success()

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        if isinstance(failure, Length.Incorrect):
            if self.minimum is None and self.maximum is not None:
                return f"Must be shorter than {self.maximum} characters."
            elif self.minimum is not None and self.maximum is None:
                return f"Must be longer than {self.minimum} characters."
            else:
                return f"Must be between {self.minimum} and {self.maximum} characters."
        return None


class Function(Validator):
    """A flexible validator which allows you to provide custom validation logic."""

    def __init__(
        self,
        function: Callable[[str], bool],
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)
        self.function = function
        """Function which takes the value to validate and returns True if valid, and False otherwise."""

    class ReturnedFalse(Failure):
        """Indicates validation failed because the supplied function returned False."""

    def validate(self, value: str) -> ValidationResult:
        """Validate that the supplied function returns True.

        Args:
            value: The value to pass into the supplied function.

        Returns:
            A ValidationResult indicating success if the function returned True,
                and failure if the function return False.
        """
        is_valid = self.function(value)
        if is_valid:
            return self.success()
        return self.failure(failures=Function.ReturnedFalse(self, value))

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        return self.failure_description


class URL(Validator):
    """Validator that checks if a URL is valid (ensuring a scheme is present)."""

    class InvalidURL(Failure):
        """Indicates that the URL is not valid."""

    def validate(self, value: str) -> ValidationResult:
        """Validates that `value` is a valid URL (contains a scheme).

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        invalid_url = ValidationResult.failure([URL.InvalidURL(self, value)])
        try:
            parsed_url = urlparse(value)
            if not all([parsed_url.scheme, parsed_url.netloc]):
                return invalid_url
        except ValueError:
            return invalid_url

        return self.success()

    def describe_failure(self, failure: Failure) -> str | None:
        """Describes why the validator failed.

        Args:
            failure: Information about why the validation failed.

        Returns:
            A string description of the failure.
        """
        return "Must be a valid URL."



from rich.segment import Segment
from rich.style import Style

from textual.color import Color
from textual.filter import DimFilter


def test_dim_apply():
    """Check dim filter changes color and resets dim attribute."""

    dim_filter = DimFilter()

    segments = [Segment("Hello, World!", Style.parse("dim #ffffff on #0000ff"))]

    dimmed_segments = dim_filter.apply(segments, Color(0, 0, 0))

    expected = [Segment("Hello, World!", Style.parse("not dim #7f7fff on #0000ff"))]

    assert dimmed_segments == expected



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from textual.app import App
    from textual.widget import Widget


def compose(node: App | Widget) -> list[Widget]:
    """Compose child widgets.

    Args:
        node: The parent node.

    Returns:
        A list of widgets.
    """
    _rich_traceback_omit = True
    from textual.widget import MountError, Widget

    app = node.app
    nodes: list[Widget] = []
    compose_stack: list[Widget] = []
    composed: list[Widget] = []
    app._compose_stacks.append(compose_stack)
    app._composed.append(composed)
    iter_compose = iter(node.compose())
    is_generator = hasattr(iter_compose, "throw")
    try:
        while True:
            try:
                child = next(iter_compose)
            except StopIteration:
                break

            if not isinstance(child, Widget):
                mount_error = MountError(
                    f"Can't mount {type(child)}; expected a Widget instance."
                )
                if is_generator:
                    iter_compose.throw(mount_error)  # type: ignore
                else:
                    raise mount_error from None

            try:
                child.id
            except AttributeError:
                mount_error = MountError(
                    "Widget is missing an 'id' attribute; did you forget to call super().__init__()?"
                )
                if is_generator:
                    iter_compose.throw(mount_error)  # type: ignore
                else:
                    raise mount_error from None

            if composed:
                nodes.extend(composed)
                composed.clear()
            if compose_stack:
                try:
                    compose_stack[-1].compose_add_child(child)
                except Exception as error:
                    if is_generator:
                        # So the error is raised inside the generator
                        # This will generate a more sensible traceback for the dev
                        iter_compose.throw(error)  # type: ignore
                    else:
                        raise
            else:
                nodes.append(child)
        if composed:
            nodes.extend(composed)
            composed.clear()
    finally:
        app._compose_stacks.pop()
        app._composed.pop()
    return nodes



from textual.app import App
from textual.containers import Container, Grid
from textual.widgets import Label


def make_label_container(  # (11)!
    text: str, id: str, border_title: str, border_subtitle: str
) -> Container:
    lbl = Label(text, id=id)
    lbl.border_title = border_title
    lbl.border_subtitle = border_subtitle
    return Container(lbl)


class BorderSubTitleAlignAll(App[None]):
    CSS_PATH = "border_sub_title_align_all.tcss"

    def compose(self):
        with Grid():
            yield make_label_container(  # (1)!
                "This is the story of",
                "lbl1",
                "[b]Border [i]title[/i][/]",
                "[u][r]Border[/r] subtitle[/]",
            )
            yield make_label_container(  # (2)!
                "a Python",
                "lbl2",
                "[b red]Left, but it's loooooooooooong",
                "[reverse]Center, but it's loooooooooooong",
            )
            yield make_label_container(  # (3)!
                "developer that",
                "lbl3",
                "[b i on purple]Left[/]",
                "[r u white on black]@@@[/]",
            )
            yield make_label_container(
                "had to fill up",
                "lbl4",
                "",  # (4)!
                "[link='https://textual.textualize.io']Left[/]",  # (5)!
            )
            yield make_label_container(  # (6)!
                "nine labels", "lbl5", "Title", "Subtitle"
            )
            yield make_label_container(  # (7)!
                "and ended up redoing it",
                "lbl6",
                "Title",
                "Subtitle",
            )
            yield make_label_container(  # (8)!
                "because the first try",
                "lbl7",
                "Title, but really loooooooooong!",
                "Subtitle, but really loooooooooong!",
            )
            yield make_label_container(  # (9)!
                "had some labels",
                "lbl8",
                "Title, but really loooooooooong!",
                "Subtitle, but really loooooooooong!",
            )
            yield make_label_container(  # (10)!
                "that were too long.",
                "lbl9",
                "Title, but really loooooooooong!",
                "Subtitle, but really loooooooooong!",
            )


if __name__ == "__main__":
    app = BorderSubTitleAlignAll()
    app.run()



from math import sin

from textual.app import App, ComposeResult
from textual.widgets import Sparkline


class SparklineColorsApp(App[None]):
    CSS_PATH = "sparkline_colors.tcss"

    def compose(self) -> ComposeResult:
        nums = [abs(sin(x / 3.14)) for x in range(0, 360 * 6, 20)]
        yield Sparkline(nums, summary_function=max, id="fst")
        yield Sparkline(nums, summary_function=max, id="snd")
        yield Sparkline(nums, summary_function=max, id="trd")
        yield Sparkline(nums, summary_function=max, id="frt")
        yield Sparkline(nums, summary_function=max, id="fft")
        yield Sparkline(nums, summary_function=max, id="sxt")
        yield Sparkline(nums, summary_function=max, id="svt")
        yield Sparkline(nums, summary_function=max, id="egt")
        yield Sparkline(nums, summary_function=max, id="nnt")
        yield Sparkline(nums, summary_function=max, id="tnt")


app = SparklineColorsApp()
if __name__ == "__main__":
    app.run()



from pathlib import Path

from textual.app import App, ComposeResult
from textual.widgets import Markdown

CSS_PATH = (Path(__file__) / "../markdown_component_classes_reloading.tcss").resolve()

CSS_PATH.write_text(
    """\
.code_inline,
.em,
.strong,
.s,
.markdown-table--header,
.markdown-table--lines,
{
    color: yellow;
}
"""
)

MD = """
# This is a **header**

| col1 | col2 |
| :- | :- |
| value 1 | value 2 |

Here's some code: `from itertools import product`.
**Bold text**
_Emphasized text_
~~strikethrough~~

```py
print("Hello, world!")
```

**That** was _some_ code.
"""


class MyApp(App[None]):
    CSS_PATH = CSS_PATH

    def compose(self) -> ComposeResult:
        yield Markdown(MD)


if __name__ == "__main__":
    MyApp().run()



import asyncio

import pytest

from textual.rlock import RLock


async def test_simple_lock():
    lock = RLock()
    # Starts not locked
    assert not lock.is_locked
    # Acquire the lock
    await lock.acquire()
    assert lock.is_locked
    # Acquire a second time (should not block)
    await lock.acquire()
    assert lock.is_locked

    # Release the lock
    lock.release()
    # Should still be locked
    assert lock.is_locked
    # Release the lock
    lock.release()
    # Should be released
    assert not lock.is_locked

    # Another release is a runtime error
    with pytest.raises(RuntimeError):
        lock.release()


async def test_multiple_tasks() -> None:
    """Check RLock prevents other tasks from acquiring lock."""
    lock = RLock()

    started: list[int] = []
    done: list[int] = []

    async def test_task(n: int) -> None:
        started.append(n)
        async with lock:
            done.append(n)

    async with lock:
        assert done == []
        task1 = asyncio.create_task(test_task(1))
        assert sorted(started) == []
        task2 = asyncio.create_task(test_task(2))
        await asyncio.sleep(0)
        assert sorted(started) == [1, 2]

    await task1
    assert 1 in done
    await task2
    assert 2 in done



"""Simple version of 5x5, developed for/with Textual."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.css.query import DOMQuery
from textual.reactive import reactive
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Button, Footer, Label, Markdown

if TYPE_CHECKING:
    from typing_extensions import Final


class Help(Screen):
    """The help screen for the application."""

    BINDINGS = [("escape,space,q,question_mark", "app.pop_screen", "Close")]
    """Bindings for the help screen."""

    def compose(self) -> ComposeResult:
        """Compose the game's help.

        Returns:
            ComposeResult: The result of composing the help screen.
        """
        yield Markdown(Path(__file__).with_suffix(".md").read_text())


class WinnerMessage(Label):
    """Widget to tell the user they have won."""

    MIN_MOVES: Final = 14
    """int: The minimum number of moves you can solve the puzzle in."""

    @staticmethod
    def _plural(value: int) -> str:
        return "" if value == 1 else "s"

    def show(self, moves: int) -> None:
        """Show the winner message.

        Args:
            moves (int): The number of moves required to win.
        """
        self.update(
            "W I N N E R !\n\n\n"
            f"You solved the puzzle in {moves} move{self._plural(moves)}."
            + (
                (
                    f" It is possible to solve the puzzle in {self.MIN_MOVES}, "
                    f"you were {moves - self.MIN_MOVES} move{self._plural(moves - self.MIN_MOVES)} over."
                )
                if moves > self.MIN_MOVES
                else " Well done! That's the minimum number of moves to solve the puzzle!"
            )
        )
        self.add_class("visible")

    def hide(self) -> None:
        """Hide the winner message."""
        self.remove_class("visible")


class GameHeader(Widget):
    """Header for the game.

    Comprises of the title (``#app-title``), the number of moves ``#moves``
    and the count of how many cells are turned on (``#progress``).
    """

    moves = reactive(0)
    """int: Keep track of how many moves the player has made."""

    filled = reactive(0)
    """int: Keep track of how many cells are filled."""

    def compose(self) -> ComposeResult:
        """Compose the game header.

        Returns:
            ComposeResult: The result of composing the game header.
        """
        with Horizontal():
            yield Label(self.app.title, id="app-title")
            yield Label(id="moves")
            yield Label(id="progress")

    def watch_moves(self, moves: int):
        """Watch the moves reactive and update when it changes.

        Args:
            moves (int): The number of moves made.
        """
        self.query_one("#moves", Label).update(f"Moves: {moves}")

    def watch_filled(self, filled: int):
        """Watch the on-count reactive and update when it changes.

        Args:
            filled (int): The number of cells that are currently on.
        """
        self.query_one("#progress", Label).update(f"Filled: {filled}")


class GameCell(Button):
    """Individual playable cell in the game."""

    @staticmethod
    def at(row: int, col: int) -> str:
        """Get the ID of the cell at the given location.

        Args:
            row (int): The row of the cell.
            col (int): The column of the cell.

        Returns:
            str: A string ID for the cell.
        """
        return f"cell-{row}-{col}"

    def __init__(self, row: int, col: int) -> None:
        """Initialise the game cell.

        Args:
            row (int): The row of the cell.
            col (int): The column of the cell.
        """
        super().__init__("", id=self.at(row, col))
        self.row = row
        self.col = col


class GameGrid(Widget):
    """The main playable grid of game cells."""

    def compose(self) -> ComposeResult:
        """Compose the game grid.

        Returns:
            ComposeResult: The result of composing the game grid.
        """
        for row in range(Game.SIZE):
            for col in range(Game.SIZE):
                yield GameCell(row, col)


class Game(Screen):
    """Main 5x5 game grid screen."""

    SIZE: Final = 5
    """The size of the game grid. Clue's in the name really."""

    BINDINGS = [
        Binding("n", "new_game", "New Game"),
        Binding("question_mark", "app.push_screen('help')", "Help", key_display="?"),
        Binding("q", "app.quit", "Quit"),
        Binding("up,w,k", "navigate(-1,0)", "Move Up", False),
        Binding("down,s,j", "navigate(1,0)", "Move Down", False),
        Binding("left,a,h", "navigate(0,-1)", "Move Left", False),
        Binding("right,d,l", "navigate(0,1)", "Move Right", False),
        Binding("space", "move", "Toggle", False),
    ]
    """The bindings for the main game grid."""

    @property
    def filled_cells(self) -> DOMQuery[GameCell]:
        """DOMQuery[GameCell]: The collection of cells that are currently turned on."""
        return cast(DOMQuery[GameCell], self.query("GameCell.filled"))

    @property
    def filled_count(self) -> int:
        """int: The number of cells that are currently filled."""
        return len(self.filled_cells)

    @property
    def all_filled(self) -> bool:
        """bool: Are all the cells filled?"""
        return self.filled_count == self.SIZE * self.SIZE

    def game_playable(self, playable: bool) -> None:
        """Mark the game as playable, or not.

        Args:
            playable (bool): Should the game currently be playable?
        """
        self.query_one(GameGrid).disabled = not playable

    def cell(self, row: int, col: int) -> GameCell:
        """Get the cell at a given location.

        Args:
            row (int): The row of the cell to get.
            col (int): The column of the cell to get.

        Returns:
            GameCell: The cell at that location.
        """
        return self.query_one(f"#{GameCell.at(row,col)}", GameCell)

    def compose(self) -> ComposeResult:
        """Compose the game screen.

        Returns:
            ComposeResult: The result of composing the game screen.
        """
        yield GameHeader()
        yield GameGrid()
        yield Footer()
        yield WinnerMessage()

    def toggle_cell(self, row: int, col: int) -> None:
        """Toggle an individual cell, but only if it's in bounds.

        If the row and column would place the cell out of bounds for the
        game grid, this function call is a no-op. That is, it's safe to call
        it with an invalid cell coordinate.

        Args:
            row (int): The row of the cell to toggle.
            col (int): The column of the cell to toggle.
        """
        if 0 <= row <= (self.SIZE - 1) and 0 <= col <= (self.SIZE - 1):
            self.cell(row, col).toggle_class("filled")

    _PATTERN: Final = (-1, 1, 0, 0, 0)

    def toggle_cells(self, cell: GameCell) -> None:
        """Toggle a 5x5 pattern around the given cell.

        Args:
            cell (GameCell): The cell to toggle the cells around.
        """
        for row, col in zip(self._PATTERN, reversed(self._PATTERN)):
            self.toggle_cell(cell.row + row, cell.col + col)
        self.query_one(GameHeader).filled = self.filled_count

    def make_move_on(self, cell: GameCell) -> None:
        """Make a move on the given cell.

        All relevant cells around the given cell are toggled as per the
        game's rules.

        Args:
            cell (GameCell): The cell to make a move on
        """
        self.toggle_cells(cell)
        self.query_one(GameHeader).moves += 1
        if self.all_filled:
            self.query_one(WinnerMessage).show(self.query_one(GameHeader).moves)
            self.game_playable(False)

    def on_button_pressed(self, event: GameCell.Pressed) -> None:
        """React to a press of a button on the game grid.

        Args:
            event (GameCell.Pressed): The event to react to.
        """
        self.make_move_on(cast(GameCell, event.button))

    def action_new_game(self) -> None:
        """Start a new game."""
        self.query_one(GameHeader).moves = 0
        self.filled_cells.remove_class("filled")
        self.query_one(WinnerMessage).hide()
        middle = self.cell(self.SIZE // 2, self.SIZE // 2)
        self.toggle_cells(middle)
        self.set_focus(middle)
        self.game_playable(True)

    def action_navigate(self, row: int, col: int) -> None:
        """Navigate to a new cell by the given offsets.

        Args:
            row (int): The row of the cell to navigate to.
            col (int): The column of the cell to navigate to.
        """
        if isinstance(self.focused, GameCell):
            self.set_focus(
                self.cell(
                    (self.focused.row + row) % self.SIZE,
                    (self.focused.col + col) % self.SIZE,
                )
            )

    def action_move(self) -> None:
        """Make a move on the current cell."""
        if isinstance(self.focused, GameCell):
            self.focused.press()

    def on_mount(self) -> None:
        """Get the game started when we first mount."""
        self.action_new_game()


class FiveByFive(App[None]):
    """Main 5x5 application class."""

    CSS_PATH = "five_by_five.tcss"
    """The name of the stylesheet for the app."""

    SCREENS = {"help": Help}
    """The pre-loaded screens for the application."""

    BINDINGS = [("ctrl+d", "toggle_dark", "Toggle Dark Mode")]
    """App-level bindings."""

    TITLE = "5x5 -- A little annoying puzzle"
    """The title of the application."""

    def on_mount(self) -> None:
        """Set up the application on startup."""
        self.push_screen(Game())


if __name__ == "__main__":
    FiveByFive().run()



"""
An implementation of a classic calculator, with a layout inspired by macOS calculator.

Works like a real calculator. Click the buttons or press the equivalent keys.
"""

from decimal import Decimal

from textual import events, on
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.css.query import NoMatches
from textual.reactive import var
from textual.widgets import Button, Digits


class CalculatorApp(App):
    """A working 'desktop' calculator."""

    CSS_PATH = "calculator.tcss"

    numbers = var("0")
    show_ac = var(True)
    left = var(Decimal("0"))
    right = var(Decimal("0"))
    value = var("")
    operator = var("plus")

    # Maps button IDs on to the corresponding key name
    NAME_MAP = {
        "asterisk": "multiply",
        "slash": "divide",
        "underscore": "plus-minus",
        "full_stop": "point",
        "plus_minus_sign": "plus-minus",
        "percent_sign": "percent",
        "equals_sign": "equals",
        "minus": "minus",
        "plus": "plus",
    }

    def watch_numbers(self, value: str) -> None:
        """Called when numbers is updated."""
        self.query_one("#numbers", Digits).update(value)

    def compute_show_ac(self) -> bool:
        """Compute switch to show AC or C button"""
        return self.value in ("", "0") and self.numbers == "0"

    def watch_show_ac(self, show_ac: bool) -> None:
        """Called when show_ac changes."""
        self.query_one("#c").display = not show_ac
        self.query_one("#ac").display = show_ac

    def compose(self) -> ComposeResult:
        """Add our buttons."""
        with Container(id="calculator"):
            yield Digits(id="numbers")
            yield Button("AC", id="ac", variant="primary")
            yield Button("C", id="c", variant="primary")
            yield Button("+/-", id="plus-minus", variant="primary")
            yield Button("%", id="percent", variant="primary")
            yield Button("÷", id="divide", variant="warning")
            yield Button("7", id="number-7", classes="number")
            yield Button("8", id="number-8", classes="number")
            yield Button("9", id="number-9", classes="number")
            yield Button("×", id="multiply", variant="warning")
            yield Button("4", id="number-4", classes="number")
            yield Button("5", id="number-5", classes="number")
            yield Button("6", id="number-6", classes="number")
            yield Button("-", id="minus", variant="warning")
            yield Button("1", id="number-1", classes="number")
            yield Button("2", id="number-2", classes="number")
            yield Button("3", id="number-3", classes="number")
            yield Button("+", id="plus", variant="warning")
            yield Button("0", id="number-0", classes="number")
            yield Button(".", id="point")
            yield Button("=", id="equals", variant="warning")

    def on_key(self, event: events.Key) -> None:
        """Called when the user presses a key."""

        def press(button_id: str) -> None:
            """Press a button, should it exist."""
            try:
                self.query_one(f"#{button_id}", Button).press()
            except NoMatches:
                pass

        key = event.key
        if key.isdecimal():
            press(f"number-{key}")
        elif key == "c":
            press("c")
            press("ac")
        else:
            button_id = self.NAME_MAP.get(key)
            if button_id is not None:
                press(self.NAME_MAP.get(key, key))

    @on(Button.Pressed, ".number")
    def number_pressed(self, event: Button.Pressed) -> None:
        """Pressed a number."""
        assert event.button.id is not None
        number = event.button.id.partition("-")[-1]
        self.numbers = self.value = self.value.lstrip("0") + number

    @on(Button.Pressed, "#plus-minus")
    def plus_minus_pressed(self) -> None:
        """Pressed + / -"""
        self.numbers = self.value = str(Decimal(self.value or "0") * -1)

    @on(Button.Pressed, "#percent")
    def percent_pressed(self) -> None:
        """Pressed %"""
        self.numbers = self.value = str(Decimal(self.value or "0") / Decimal(100))

    @on(Button.Pressed, "#point")
    def pressed_point(self) -> None:
        """Pressed ."""
        if "." not in self.value:
            self.numbers = self.value = (self.value or "0") + "."

    @on(Button.Pressed, "#ac")
    def pressed_ac(self) -> None:
        """Pressed AC"""
        self.value = ""
        self.left = self.right = Decimal(0)
        self.operator = "plus"
        self.numbers = "0"

    @on(Button.Pressed, "#c")
    def pressed_c(self) -> None:
        """Pressed C"""
        self.value = ""
        self.numbers = "0"

    def _do_math(self) -> None:
        """Does the math: LEFT OPERATOR RIGHT"""
        try:
            if self.operator == "plus":
                self.left += self.right
            elif self.operator == "minus":
                self.left -= self.right
            elif self.operator == "divide":
                self.left /= self.right
            elif self.operator == "multiply":
                self.left *= self.right
            self.numbers = str(self.left)
            self.value = ""
        except Exception:
            self.numbers = "Error"

    @on(Button.Pressed, "#plus,#minus,#divide,#multiply")
    def pressed_op(self, event: Button.Pressed) -> None:
        """Pressed one of the arithmetic operations."""
        self.right = Decimal(self.value or "0")
        self._do_math()
        assert event.button.id is not None
        self.operator = event.button.id

    @on(Button.Pressed, "#equals")
    def pressed_equals(self) -> None:
        """Pressed ="""
        if self.value:
            self.right = Decimal(self.value)
        self._do_math()


if __name__ == "__main__":
    CalculatorApp().run(inline=True)



from textual.app import App, ComposeResult, RenderResult
from textual.reactive import reactive
from textual.widgets import Footer, Static


class Counter(Static, can_focus=True):
    """A counter that can be incremented and decremented by pressing keys."""

    BINDINGS = [
        ("up,k", "change_count(1)", "Increment"),  # (1)!
        ("down,j", "change_count(-1)", "Decrement"),
    ]

    count = reactive(0)

    def render(self) -> RenderResult:
        return f"Count: {self.count}"

    def action_change_count(self, amount: int) -> None:  # (2)!
        self.count += amount


class CounterApp(App[None]):
    CSS_PATH = "counter.tcss"

    def compose(self) -> ComposeResult:
        yield Counter()
        yield Counter()
        yield Counter()
        yield Footer()


if __name__ == "__main__":
    app = CounterApp()
    app.run()



"""Widgets for showing notification messages in toasts."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from rich.text import Text

from textual import on

if TYPE_CHECKING:
    from textual.app import RenderResult

from textual.containers import Container
from textual.css.query import NoMatches
from textual.events import Click, Mount
from textual.notifications import Notification, Notifications
from textual.widgets._static import Static


class ToastHolder(Container, inherit_css=False):
    """Container that holds a single toast.

    Used to control the alignment of each of the toasts in the main toast
    container.
    """

    DEFAULT_CSS = """
    ToastHolder {
        align-horizontal: right;
        width: 1fr;
        height: auto;
        visibility: hidden;
    }
    """


class Toast(Static, inherit_css=False):
    """A widget for displaying short-lived notifications."""

    DEFAULT_CSS = """
    Toast {
        width: 60;
        max-width: 50%;
        height: auto;
        visibility: visible;
        margin-top: 1;
        padding: 1 1;
        background: $panel-lighten-1;
        link-background: initial;
        link-color: $foreground;
        link-style: underline;
        link-background-hover: $primary;
        link-color-hover: $foreground;
        link-style-hover: bold not underline;
    }

    .toast--title {
        text-style: bold;
        color: $foreground;
    }

    Toast.-information {
        border-left: outer $success;
    }

    Toast.-information .toast--title {
        color: $text-success;
    }

    Toast.-warning {
        border-left: outer $warning;
    }

    Toast.-warning .toast--title {
        color: $text-warning;
    }

    Toast.-error {
        border-left: outer $error;
    }

    Toast.-error .toast--title {
       color: $text-error;
    }
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {"toast--title"}
    """
    | Class | Description |
    | :- | :- |
    | `toast--title` | Targets the title of the toast. |
    """

    DEFAULT_CLASSES = "-textual-system"

    def __init__(self, notification: Notification) -> None:
        """Initialise the toast.

        Args:
            notification: The notification to show in the toast.
        """
        super().__init__(classes=f"-{notification.severity}")
        self._notification = notification
        self._timeout = notification.time_left

    def render(self) -> RenderResult:
        """Render the toast's content.

        Returns:
            A Rich renderable for the title and content of the Toast.
        """
        notification = self._notification
        if notification.title:
            header_style = self.get_component_rich_style("toast--title")
            notification_text = Text.assemble(
                (notification.title, header_style),
                "\n",
                Text.from_markup(notification.message),
            )
        else:
            notification_text = Text.assemble(
                Text.from_markup(notification.message),
            )
        return notification_text

    def _on_mount(self, _: Mount) -> None:
        """Set the time running once the toast is mounted."""
        self.set_timer(self._timeout, self._expire)

    @on(Click)
    def _expire(self) -> None:
        """Remove the toast once the timer has expired."""
        # Before we removed ourself, we also call on the app to forget about
        # the notification that caused us to exist. Note that we tell the
        # app to not bother refreshing the display on our account, we're
        # about to handle that anyway.
        self.app._unnotify(self._notification, refresh=False)
        # Note that we attempt to remove our parent, because we're wrapped
        # inside an alignment container. The testing that we are is as much
        # to keep type checkers happy as anything else.
        (self.parent if isinstance(self.parent, ToastHolder) else self).remove()


class ToastRack(Container, inherit_css=False):
    """A container for holding toasts."""

    DEFAULT_CSS = """
    ToastRack {
        layer: _toastrack;
        width: 1fr;
        height: auto;
        dock: bottom;
        align: right bottom;
        visibility: hidden;
        layout: vertical;
        overflow-y: scroll;
        margin-bottom: 1;
    }
    """
    DEFAULT_CLASSES = "-textual-system"

    @staticmethod
    def _toast_id(notification: Notification) -> str:
        """Create a Textual-DOM-internal ID for the given notification.

        Args:
            notification: The notification to create the ID for.

        Returns:
            An ID for the notification that can be used within the DOM.
        """
        return f"--textual-toast-{notification.identity}"

    def show(self, notifications: Notifications) -> None:
        """Show the notifications as toasts.

        Args:
            notifications: The notifications to show.
        """
        # Look for any stale toasts and remove them.
        for toast in self.query(Toast):
            if toast._notification not in notifications:
                toast.remove()

        # Gather up all the notifications that we don't have toasts for yet.
        new_toasts: list[Notification] = []
        for notification in notifications:
            try:
                # See if there's already a toast for that notification.
                _ = self.get_child_by_id(self._toast_id(notification))
            except NoMatches:
                if not notification.has_expired:
                    new_toasts.append(notification)

        # If we got any...
        if new_toasts:
            # ...mount them.
            self.mount_all(
                ToastHolder(Toast(toast), id=self._toast_id(toast))
                for toast in new_toasts
            )
            self.call_later(self.scroll_end, animate=False, force=True)



from __future__ import annotations

from dataclasses import dataclass, field
from functools import partial
from typing import Callable

from textual.command import DiscoveryHit, Hit, Hits, Provider
from textual.design import ColorSystem


@dataclass
class Theme:
    """Defines a theme for the application."""

    name: str
    """The name of the theme.

    After registering a theme with `App.register_theme`, you can set the theme with
    `App.theme = theme_name`. This will immediately apply the theme's colors to your
    application.
    """

    primary: str
    secondary: str | None = None
    warning: str | None = None
    error: str | None = None
    success: str | None = None
    accent: str | None = None
    foreground: str | None = None
    background: str | None = None
    surface: str | None = None
    panel: str | None = None
    boost: str | None = None
    dark: bool = True
    luminosity_spread: float = 0.15
    text_alpha: float = 0.95
    variables: dict[str, str] = field(default_factory=dict)

    def to_color_system(self) -> ColorSystem:
        """
        Create a ColorSystem instance from this Theme.

        Returns:
            A ColorSystem instance with attributes copied from this Theme.
        """
        return ColorSystem(
            primary=self.primary,
            secondary=self.secondary,
            warning=self.warning,
            error=self.error,
            success=self.success,
            accent=self.accent,
            foreground=self.foreground,
            background=self.background,
            surface=self.surface,
            panel=self.panel,
            boost=self.boost,
            dark=self.dark,
            luminosity_spread=self.luminosity_spread,
            text_alpha=self.text_alpha,
            variables=self.variables,
        )


BUILTIN_THEMES: dict[str, Theme] = {
    "textual-dark": Theme(
        name="textual-dark",
        primary="#0178D4",
        secondary="#004578",
        accent="#ffa62b",
        warning="#ffa62b",
        error="#ba3c5b",
        success="#4EBF71",
        foreground="#e0e0e0",
    ),
    "textual-light": Theme(
        name="textual-light",
        primary="#004578",
        secondary="#0178D4",
        accent="#ffa62b",
        warning="#ffa62b",
        error="#ba3c5b",
        success="#4EBF71",
        surface="#D8D8D8",
        panel="#D0D0D0",
        background="#E0E0E0",
        dark=False,
        variables={
            "footer-key-foreground": "#0178D4",
        },
    ),
    "nord": Theme(
        name="nord",
        primary="#88C0D0",
        secondary="#81A1C1",
        accent="#B48EAD",
        foreground="#D8DEE9",
        background="#2E3440",
        success="#A3BE8C",
        warning="#EBCB8B",
        error="#BF616A",
        surface="#3B4252",
        panel="#434C5E",
        variables={
            "block-cursor-background": "#88C0D0",
            "block-cursor-foreground": "#2E3440",
            "block-cursor-text-style": "none",
            "footer-key-foreground": "#88C0D0",
            "input-selection-background": "#81a1c1 35%",
            "button-color-foreground": "#2E3440",
            "button-focus-text-style": "reverse",
        },
    ),
    "gruvbox": Theme(
        name="gruvbox",
        primary="#85A598",
        secondary="#A89A85",
        warning="#fe8019",
        error="#fb4934",
        success="#b8bb26",
        accent="#fabd2f",
        foreground="#fbf1c7",
        background="#282828",
        surface="#3c3836",
        panel="#504945",
        variables={
            "block-cursor-foreground": "#fbf1c7",
            "input-selection-background": "#689d6a40",
            "button-color-foreground": "#282828",
        },
    ),
    "catppuccin-mocha": Theme(
        name="catppuccin-mocha",
        primary="#F5C2E7",
        secondary="#cba6f7",
        warning="#FAE3B0",
        error="#F28FAD",
        success="#ABE9B3",
        accent="#fab387",
        foreground="#cdd6f4",
        background="#181825",
        surface="#313244",
        panel="#45475a",
        variables={
            "input-cursor-foreground": "#11111b",
            "input-cursor-background": "#f5e0dc",
            "input-selection-background": "#9399b2 30%",
            "border": "#b4befe",
            "border-blurred": "#585b70",
            "footer-background": "#45475a",
            "block-cursor-foreground": "#1e1e2e",
            "block-cursor-text-style": "none",
            "button-color-foreground": "#181825",
        },
    ),
    "textual-ansi": Theme(
        name="textual-ansi",
        primary="ansi_blue",
        secondary="ansi_cyan",
        warning="ansi_yellow",
        error="ansi_red",
        success="ansi_green",
        accent="ansi_bright_blue",
        foreground="ansi_default",
        background="ansi_default",
        surface="ansi_default",
        panel="ansi_default",
        boost="ansi_default",
        dark=False,
        variables={
            "block-cursor-text-style": "b",
            "block-cursor-blurred-text-style": "i",
            "input-selection-background": "ansi_blue",
            "input-cursor-text-style": "reverse",
            "scrollbar": "ansi_blue",
            "border-blurred": "ansi_blue",
            "border": "ansi_bright_blue",
        },
    ),
    "dracula": Theme(
        name="dracula",
        primary="#BD93F9",
        secondary="#6272A4",
        warning="#FFB86C",
        error="#FF5555",
        success="#50FA7B",
        accent="#FF79C6",
        background="#282A36",
        surface="#2B2E3B",
        panel="#313442",
        foreground="#F8F8F2",
        variables={
            "button-color-foreground": "#282A36",
        },
    ),
    "tokyo-night": Theme(
        name="tokyo-night",
        primary="#BB9AF7",
        secondary="#7AA2F7",
        warning="#E0AF68",  # Yellow
        error="#F7768E",  # Red
        success="#9ECE6A",  # Green
        accent="#FF9E64",  # Orange
        foreground="#a9b1d6",
        background="#1A1B26",  # Background
        surface="#24283B",  # Surface
        panel="#414868",  # Panel
        variables={
            "button-color-foreground": "#24283B",
        },
    ),
    "monokai": Theme(
        name="monokai",
        primary="#AE81FF",
        secondary="#F92672",
        accent="#66D9EF",
        warning="#FD971F",
        error="#F92672",
        success="#A6E22E",
        foreground="#d6d6d6",
        background="#272822",
        surface="#2e2e2e",
        panel="#3E3D32",
        variables={
            "foreground-muted": "#797979",
            "input-selection-background": "#575b6190",
            "button-color-foreground": "#272822",
        },
    ),
    "flexoki": Theme(
        name="flexoki",
        primary="#205EA6",  # blue
        secondary="#24837B",  # cyan
        warning="#AD8301",  # yellow
        error="#AF3029",  # red
        success="#66800B",  # green
        accent="#9B76C8",  # purple light
        background="#100F0F",  # base.black
        surface="#1C1B1A",  # base.950
        panel="#282726",  # base.900
        foreground="#FFFCF0",  # base.paper
        variables={
            "input-cursor-foreground": "#5E409D",
            "input-cursor-background": "#FFFCF0",
            "input-selection-background": "#6F6E69 35%",  # base.600 with opacity
            "button-color-foreground": "#FFFCF0",
        },
    ),
    "catppuccin-latte": Theme(
        name="catppuccin-latte",
        secondary="#DC8A78",
        primary="#8839EF",
        warning="#DF8E1D",
        error="#D20F39",
        success="#40A02B",
        accent="#FE640B",
        foreground="#4C4F69",
        background="#EFF1F5",
        surface="#E6E9EF",
        panel="#CCD0DA",
        dark=False,
        variables={
            "button-color-foreground": "#EFF1F5",
        },
    ),
    "solarized-light": Theme(
        name="solarized-light",
        primary="#268bd2",
        secondary="#2aa198",
        warning="#cb4b16",
        error="#dc322f",
        success="#859900",
        accent="#6c71c4",
        foreground="#586e75",
        background="#fdf6e3",
        surface="#eee8d5",
        panel="#eee8d5",
        dark=False,
        variables={
            "button-color-foreground": "#fdf6e3",
            "footer-background": "#268bd2",
            "footer-key-foreground": "#fdf6e3",
            "footer-description-foreground": "#fdf6e3",
        },
    ),
}


class ThemeProvider(Provider):
    """A provider for themes."""

    @property
    def commands(self) -> list[tuple[str, Callable[[], None]]]:
        themes = self.app.available_themes

        def set_app_theme(name: str) -> None:
            self.app.theme = name

        return [
            (theme.name, partial(set_app_theme, theme.name))
            for theme in themes.values()
            if theme.name != "textual-ansi"
        ]

    async def discover(self) -> Hits:
        for command in self.commands:
            yield DiscoveryHit(*command)

    async def search(self, query: str) -> Hits:
        matcher = self.matcher(query)

        for name, callback in self.commands:
            if (match := matcher.match(name)) > 0:
                yield Hit(
                    match,
                    matcher.highlight(name),
                    callback,
                )



from textual.app import App, ComposeResult
from textual.clock import MockClock
from textual.containers import Center, Middle
from textual.timer import Timer
from textual.widgets import Footer, ProgressBar


class StyledProgressBar(App[None]):
    BINDINGS = [("s", "start", "Start")]
    CSS_PATH = "progress_bar_styled.tcss"

    progress_timer: Timer
    """Timer to simulate progress happening."""

    def compose(self) -> ComposeResult:
        self.clock = MockClock()
        with Center():
            with Middle():
                yield ProgressBar(clock=self.clock)
        yield Footer()

    def on_mount(self) -> None:
        """Set up a timer to simulate progress happening."""
        self.progress_timer = self.set_interval(1 / 10, self.make_progress, pause=True)

    def make_progress(self) -> None:
        """Called automatically to advance the progress bar."""
        self.query_one(ProgressBar).advance(1)

    def action_start(self) -> None:
        """Start the progress tracking."""
        self.query_one(ProgressBar).update(total=100)
        self.progress_timer.resume()
        self.query_one(ProgressBar).refresh()

    def key_f(self) -> None:
        # Freeze time for the indeterminate progress bar.
        self.clock.set_time(5.0)

    def key_t(self) -> None:
        # Freeze time to show always the same ETA.
        self.clock.set_time(0)
        self.query_one(ProgressBar).update(total=100, progress=0)
        self.clock.set_time(3.9)
        self.query_one(ProgressBar).update(progress=39)

    def key_u(self) -> None:
        self.query_one(ProgressBar).update(total=100, progress=100)


if __name__ == "__main__":
    StyledProgressBar().run()



from textual.app import App, ComposeResult
from textual.events import AppBlur
from textual.widgets import Input

class AppBlurApp(App[None]):

    CSS = """
    Screen {
        align: center middle;
    }

    Input {
        width: 50%;
        margin-bottom: 1;

        &:focus {
            width: 75%;
            border: thick green;
            background: pink;
        }
    }
    """

    def compose(self) -> ComposeResult:
        yield Input("This should be the blur style")
        yield Input("This should also be the blur style")

    def on_mount(self) -> None:
        self.post_message(AppBlur())

if __name__ == "__main__":
    AppBlurApp().run()



from itertools import cycle

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.widgets import Footer, Header, Input, Label, SelectionList


class TodoList(App[None]):
    CSS = """
Screen {
    align: center middle;
    hatch: right $foreground 10%;
}
#content {
    height: auto;
    width: 40;
    padding: 1 2;
}
#header {
    height: 1;
    width: auto;
    margin-bottom: 1;
}
.title {
    text-style: bold;
    padding: 0 1;
    width: 1fr;
}
#overdue {
    color: $text-error;
    background: $error-muted;
    padding: 0 1;
    width: auto;
}
#done {
    color: $text-success;
    background: $success-muted;
    padding: 0 1;
    margin: 0 1;
}
#footer {
    height: auto;
    margin-bottom: 2;
}
#history-header {
    height: 1;
    width: auto;
}
#history-done {
    width: auto;
    padding: 0 1;
    margin: 0 1;
    background: $primary-muted;
    color: $text-primary;
}
"""

    BINDINGS = [Binding("ctrl+t", "cycle_theme", "Cycle theme")]
    THEMES = cycle(
        ["nord", "gruvbox", "tokyo-night", "textual-dark", "solarized-light"]
    )

    def compose(self) -> ComposeResult:
        yield Header()
        with Vertical(id="content"):
            with Horizontal(id="header"):
                yield Label("Today", classes="title")
                yield Label("1 overdue", id="overdue")
                yield Label("1 done", id="done")
            yield SelectionList(
                ("Buy milk", 0),
                ("Buy bread", 1),
                ("Go and vote", 2, True),
                ("Return package", 3),
                id="todo-list",
            )
            with Horizontal(id="footer"):
                yield Input(placeholder="Add a task")

            with Horizontal(id="history-header"):
                yield Label("History", classes="title")
                yield Label("4 items", id="history-done")

        yield Footer()

    def on_mount(self) -> None:
        self.action_cycle_theme()

    def action_cycle_theme(self) -> None:
        self.theme = next(self.THEMES)


app = TodoList()
if __name__ == "__main__":
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Markdown

EXAMPLE_MARKDOWN = """\
# Markdown Document

This is an example of Textual's `Markdown` widget.

## Features

Markdown syntax and extensions are supported.

- Typography *emphasis*, **strong**, `inline code` etc.
- Headers
- Lists (bullet and ordered)
- Syntax highlighted code blocks
- Tables!
"""


class MarkdownExampleApp(App):
    def compose(self) -> ComposeResult:
        yield Markdown(EXAMPLE_MARKDOWN)


if __name__ == "__main__":
    app = MarkdownExampleApp()
    app.run()



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widget import Widget
from textual.widgets import Input, Label, Switch


class BitSwitch(Widget):
    """A Switch with a numeric label above it."""

    DEFAULT_CSS = """
    BitSwitch {
        layout: vertical;
        width: auto;
        height: auto;
    }
    BitSwitch > Label {
        text-align: center;
        width: 100%;
    }
    """

    def __init__(self, bit: int) -> None:
        self.bit = bit
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label(str(self.bit))
        yield Switch()


class ByteInput(Widget):
    """A compound widget with 8 switches."""

    DEFAULT_CSS = """
    ByteInput {
        width: auto;
        height: auto;
        border: blank;
        layout: horizontal;
    }
    ByteInput:focus-within {
        border: heavy $secondary;
    }
    """

    def compose(self) -> ComposeResult:
        for bit in reversed(range(8)):
            yield BitSwitch(bit)


class ByteEditor(Widget):
    DEFAULT_CSS = """
    ByteEditor > Container {
        height: 1fr;
        align: center middle;
    }
    ByteEditor > Container.top {
        background: $boost;
    }
    ByteEditor Input {
        width: 16;
    }
    """

    def compose(self) -> ComposeResult:
        with Container(classes="top"):
            yield Input(placeholder="byte")
        with Container():
            yield ByteInput()


class ByteInputApp(App):
    def compose(self) -> ComposeResult:
        yield ByteEditor()


if __name__ == "__main__":
    app = ByteInputApp()
    app.run()



from __future__ import annotations

from dataclasses import dataclass

import pytest

from textual import on
from textual._on import OnDecoratorError
from textual.app import App, ComposeResult
from textual.events import Enter, Leave
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Button, TabbedContent, TabPane


async def test_on_button_pressed() -> None:
    """Test handlers with @on decorator."""

    pressed: list[str] = []

    class ButtonApp(App):
        def compose(self) -> ComposeResult:
            yield Button("OK", id="ok")
            yield Button("Cancel", classes="exit cancel")
            yield Button("Quit", classes="exit quit")

        @on(Button.Pressed, "#ok")
        def ok(self):
            pressed.append("ok")

        @on(Button.Pressed, ".exit")
        def exit(self):
            pressed.append("exit")

        @on(Button.Pressed, ".exit.quit")
        def _(self):
            pressed.append("quit")

        def on_button_pressed(self):
            pressed.append("default")

    app = ButtonApp()
    async with app.run_test() as pilot:
        await pilot.press("enter", "tab", "enter", "tab", "enter")
        await pilot.pause()

    assert pressed == [
        "ok",  # Matched ok first
        "default",  # on_button_pressed matched everything
        "exit",  # Cancel button, matches exit
        "default",  # on_button_pressed matched everything
        "exit",  # Quit button pressed, matched exit and _
        "quit",  # Matched previous button
        "default",  # on_button_pressed matched everything
    ]


async def test_on_inheritance() -> None:
    """Test on decorator and inheritance."""
    pressed: list[str] = []

    class MyWidget(Widget):
        def compose(self) -> ComposeResult:
            yield Button("OK", id="ok")

        # Also called
        @on(Button.Pressed, "#ok")
        def ok(self):
            pressed.append("MyWidget.ok base")

    class DerivedWidget(MyWidget):
        # Should be called first
        @on(Button.Pressed, "#ok")
        def ok(self):
            pressed.append("MyWidget.ok derived")

    class ButtonApp(App):
        def compose(self) -> ComposeResult:
            yield DerivedWidget()

    app = ButtonApp()
    async with app.run_test() as pilot:
        await pilot.press("tab", "enter")

    expected = ["MyWidget.ok derived", "MyWidget.ok base"]
    assert pressed == expected


def test_on_bad_selector() -> None:
    """Check bad selectors raise an error."""

    with pytest.raises(OnDecoratorError):

        @on(Button.Pressed, "@")
        def foo():
            pass


def test_on_no_control() -> None:
    """Check messages with no 'control' attribute raise an error."""

    class CustomMessage(Message):
        pass

    with pytest.raises(OnDecoratorError):

        @on(CustomMessage, "#foo")
        def foo():
            pass


def test_on_attribute_not_listed() -> None:
    """Check `on` checks if the attribute is in ALLOW_SELECTOR_MATCH."""

    class CustomMessage(Message):
        pass

    with pytest.raises(OnDecoratorError):

        @on(CustomMessage, foo="bar")
        def foo():
            pass


async def test_on_arbitrary_attributes() -> None:
    log: list[str] = []

    class OnArbitraryAttributesApp(App[None]):
        def compose(self) -> ComposeResult:
            with TabbedContent():
                yield TabPane("One", id="one")
                yield TabPane("Two", id="two")
                yield TabPane("Three", id="three")

        def on_mount(self) -> None:
            self.query_one(TabbedContent).add_class("tabs")

        @on(TabbedContent.TabActivated, pane="#one")
        def one(self) -> None:
            log.append("one")

        @on(TabbedContent.TabActivated, pane="#two")
        def two(self) -> None:
            log.append("two")

    app = OnArbitraryAttributesApp()
    async with app.run_test() as pilot:
        await pilot.press("tab", "right", "right")

    assert log == ["one", "two"]


class MessageSender(Widget):
    @dataclass
    class Parent(Message):
        sender: MessageSender

        @property
        def control(self) -> MessageSender:
            return self.sender

    class Child(Parent):
        pass

    def post_parent(self) -> None:
        self.post_message(self.Parent(self))

    def post_child(self) -> None:
        self.post_message(self.Child(self))


async def test_fire_on_inherited_message() -> None:
    """Handlers should fire when descendant messages are posted."""

    posted: list[str] = []

    class InheritTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MessageSender()

        @on(MessageSender.Parent)
        def catch_parent(self) -> None:
            posted.append("parent")

        @on(MessageSender.Child)
        def catch_child(self) -> None:
            posted.append("child")

        def on_mount(self) -> None:
            self.query_one(MessageSender).post_parent()
            self.query_one(MessageSender).post_child()

    async with InheritTestApp().run_test():
        pass

    assert posted == ["parent", "child", "parent"]


async def test_fire_inherited_on_single_handler() -> None:
    """Test having parent/child messages on a single handler."""

    posted: list[str] = []

    class InheritTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MessageSender()

        @on(MessageSender.Parent)
        @on(MessageSender.Child)
        def catch_either(self, event: MessageSender.Parent) -> None:
            posted.append(f"either {event.__class__.__name__}")

        def on_mount(self) -> None:
            self.query_one(MessageSender).post_parent()
            self.query_one(MessageSender).post_child()

    async with InheritTestApp().run_test():
        pass

    assert posted == ["either Parent", "either Child"]


async def test_fire_inherited_on_single_handler_multi_selector() -> None:
    """Test having parent/child messages on a single handler but with different selectors."""

    posted: list[str] = []

    class InheritTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MessageSender(classes="a b")

        @on(MessageSender.Parent, ".y")
        @on(MessageSender.Child, ".y")
        @on(MessageSender.Parent, ".a.b")
        @on(MessageSender.Child, ".a.b")
        @on(MessageSender.Parent, ".a")
        @on(MessageSender.Child, ".a")
        @on(MessageSender.Parent, ".b")
        @on(MessageSender.Child, ".b")
        @on(MessageSender.Parent, ".x")
        @on(MessageSender.Child, ".x")
        def catch_either(self, event: MessageSender.Parent) -> None:
            posted.append(f"either {event.__class__.__name__}")

        @on(MessageSender.Child, ".a, .x")
        def catch_selector_list_one_miss(self, event: MessageSender.Parent) -> None:
            posted.append(f"selector list one miss {event.__class__.__name__}")

        @on(MessageSender.Child, ".a, .b")
        def catch_selector_list_two_hits(self, event: MessageSender.Parent) -> None:
            posted.append(f"selector list two hits {event.__class__.__name__}")

        @on(MessageSender.Child, ".a.b")
        def catch_selector_combined_hits(self, event: MessageSender.Parent) -> None:
            posted.append(f"combined hits {event.__class__.__name__}")

        @on(MessageSender.Child, ".a.x")
        def catch_selector_combined_miss(self, event: MessageSender.Parent) -> None:
            posted.append(f"combined miss {event.__class__.__name__}")

        def on_mount(self) -> None:
            self.query_one(MessageSender).post_parent()
            self.query_one(MessageSender).post_child()

    async with InheritTestApp().run_test():
        pass

    assert posted == [
        "either Parent",
        "either Child",
        "selector list one miss Child",
        "selector list two hits Child",
        "combined hits Child",
    ]


async def test_fire_inherited_and_on_methods() -> None:
    posted: list[str] = []

    class OnAndOnTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MessageSender()

        def on_message_sender_parent(self) -> None:
            posted.append("on_message_sender_parent")

        @on(MessageSender.Parent)
        def catch_parent(self) -> None:
            posted.append("catch_parent")

        def on_message_sender_child(self) -> None:
            posted.append("on_message_sender_child")

        @on(MessageSender.Child)
        def catch_child(self) -> None:
            posted.append("catch_child")

        def on_mount(self) -> None:
            self.query_one(MessageSender).post_parent()
            self.query_one(MessageSender).post_child()

    async with OnAndOnTestApp().run_test():
        pass

    assert posted == [
        "catch_parent",
        "on_message_sender_parent",
        "catch_child",
        "catch_parent",
        "on_message_sender_child",
    ]


class MixinMessageSender(Widget):
    class Parent(Message):
        pass

    class JustSomeRandomMixin:
        pass

    class Child(JustSomeRandomMixin, Parent):
        pass

    def post_parent(self) -> None:
        self.post_message(self.Parent())

    def post_child(self) -> None:
        self.post_message(self.Child())


async def test_fire_on_inherited_message_plus_mixins() -> None:
    """Handlers should fire when descendant messages are posted, without mixins messing things up."""

    posted: list[str] = []

    class InheritTestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield MixinMessageSender()

        @on(MixinMessageSender.Parent)
        def catch_parent(self) -> None:
            posted.append("parent")

        @on(MixinMessageSender.Child)
        def catch_child(self) -> None:
            posted.append("child")

        def on_mount(self) -> None:
            self.query_one(MixinMessageSender).post_parent()
            self.query_one(MixinMessageSender).post_child()

    async with InheritTestApp().run_test():
        pass

    assert posted == ["parent", "child", "parent"]


async def test_on_with_enter_and_leave_events():
    class EnterLeaveApp(App):
        messages = []

        def compose(self) -> ComposeResult:
            yield Button("OK")

        @on(Enter, "Button")
        @on(Leave, "Button")
        def record(self, event: Enter | Leave) -> None:
            self.messages.append(event.__class__.__name__)

    app = EnterLeaveApp()
    async with app.run_test() as pilot:
        expected_messages = []

        await pilot.hover(Button)
        expected_messages.append("Enter")
        assert app.messages == expected_messages

        await pilot.hover(Button, offset=(0, 20))
        expected_messages.append("Leave")
        assert app.messages == expected_messages



from time import time

from textual.app import App, ComposeResult, RenderableType
from textual.containers import Container
from textual.renderables.gradient import LinearGradient
from textual.widgets import Static

COLORS = [
    "#881177",
    "#aa3355",
    "#cc6666",
    "#ee9944",
    "#eedd00",
    "#99dd55",
    "#44dd88",
    "#22ccbb",
    "#00bbcc",
    "#0099cc",
    "#3366bb",
    "#663399",
]
STOPS = [(i / (len(COLORS) - 1), color) for i, color in enumerate(COLORS)]


class Splash(Container):
    """Custom widget that extends Container."""

    DEFAULT_CSS = """
    Splash {
        align: center middle;
    }
    Static {
        width: 40;
        padding: 2 4;
    }
    """

    def on_mount(self) -> None:
        self.auto_refresh = 1 / 30

    def compose(self) -> ComposeResult:
        yield Static("Making a splash with Textual!")

    def render(self) -> RenderableType:
        return LinearGradient(time() * 90, STOPS)


class SplashApp(App):
    """Simple app to show our custom widget."""

    def compose(self) -> ComposeResult:
        yield Splash()


if __name__ == "__main__":
    app = SplashApp()
    app.run()
    print("https://textual.textualize.io/how-to/render-and-compose/")



from textual.app import App, ComposeResult
from textual.widgets import Label, TabbedContent, TabPane


class TabApp(App):
    CSS = """
    TabPane {
    border: solid blue;
    }
    """

    def compose(self) -> ComposeResult:
        with TabbedContent():
            with TabPane("Tab 1", id="tab-1"):
                yield Label("hello")
                yield Label("hello")
                yield Label("hello")
            with TabPane("Tab 2", id="tab-2"):
                yield Label("world")


if __name__ == "__main__":
    app = TabApp()
    app.run()



from textual import __version__
from textual.app import App, ComposeResult
from textual.widgets import Static


# Regression test for https://github.com/Textualize/textual/issues/3858
class BrokenClassesApp(App[None]):
    CSS = """
    Screen {
        align: center middle;
    }

    Static {
        width: 50%;
        height: 50%;
        border: solid;
    }

    Screen.go-red Static {
        background: red;
    }
    """

    def compose(self) -> ComposeResult:
        yield Static("This should have a red background")

    def on_mount(self) -> None:
        self.screen.set_class(True, "go-red")


if __name__ == "__main__":
    BrokenClassesApp().run()



from rich.text import Text

from textual.app import App
from textual.widgets import Select, Static
from textual.widgets._select import SelectCurrent, SelectOverlay


async def test_reactive_prompt_change():
    """Regression test for https://github.com/Textualize/textual/issues/2983"""

    class SelectApp(App):
        def compose(self):
            yield Select[int](
                [(str(n), n) for n in range(3)],
                prompt="Old prompt",
            )

    app = SelectApp()
    async with app.run_test() as pilot:
        select_widget = pilot.app.query_one(Select)
        select_current = select_widget.query_one(SelectCurrent)
        select_current_label = select_current.query_one("#label", Static)
        select_overlay = select_widget.query_one(SelectOverlay)

        assert select_current_label.renderable == "Old prompt"
        assert select_overlay._options[0].prompt == Text("Old prompt")

        select_widget.prompt = "New prompt"
        assert select_current_label.renderable == "New prompt"
        assert select_overlay._options[0].prompt == Text("New prompt")


async def test_reactive_prompt_change_when_allow_blank_is_false():
    class SelectApp(App):
        def compose(self):
            yield Select[int](
                [(str(n), n) for n in range(3)],
                prompt="Old prompt",
                allow_blank=False,
            )

    app = SelectApp()
    async with app.run_test() as pilot:
        select_widget = pilot.app.query_one(Select)
        select_current = select_widget.query_one(SelectCurrent)
        select_current_label = select_current.query_one("#label", Static)
        select_overlay = select_widget.query_one(SelectOverlay)

        assert select_current_label.renderable == "0"
        assert select_overlay._options[0].prompt == "0"

        select_widget.prompt = "New prompt"
        assert select_current_label.renderable == "0"
        assert select_overlay._options[0].prompt == "0"



from textual.app import App
from textual.color import Color
from textual.screen import Screen
from textual.widgets import Label

RED = Color(255, 0, 0)
GREEN = Color(0, 255, 0)
BLUE = Color(0, 0, 255)


class BaseScreen(Screen):
    def compose(self):
        yield Label("Hello, world!", id="app-css")
        yield Label("Hello, world!", id="screen-css-path")
        yield Label("Hello, world!", id="screen-css")


class ScreenWithCSS(Screen):
    SCOPED_CSS = False
    CSS = """
    #screen-css {
        background: #ff0000;
    }
    """

    CSS_PATH = "test_screen_css.tcss"

    def compose(self):
        yield Label("Hello, world!", id="app-css")
        yield Label("Hello, world!", id="screen-css-path")
        yield Label("Hello, world!", id="screen-css")


class BaseApp(App):
    """Base app for testing screen CSS when pushing screens."""

    CSS = """
    #app-css {
        background: #00ff00;
    }
    #screen-css-path {
        background: #00ff00;
    }
    #screen-css {
        background: #00ff00;
    }
    """

    def on_mount(self):
        self.push_screen(BaseScreen())


class SwitchBaseApp(BaseApp):
    """Base app for testing screen CSS when switching a screen."""

    def on_mount(self):
        self.push_screen(BaseScreen())


def check_colors_before_screen_css(app: BaseApp):
    assert app.query_one("#app-css").styles.background == GREEN
    assert app.query_one("#screen-css-path").styles.background == GREEN
    assert app.query_one("#screen-css").styles.background == GREEN


def check_colors_after_screen_css(app: BaseApp):
    assert app.query_one("#app-css").styles.background == GREEN
    assert app.query_one("#screen-css-path").styles.background == BLUE
    assert app.query_one("#screen-css").styles.background == RED


async def test_screen_pushing_and_popping_does_not_reparse_css():
    """Check that pushing and popping the same screen doesn't trigger CSS reparses."""

    class MyApp(BaseApp):
        def key_p(self):
            self.push_screen(ScreenWithCSS())

        def key_o(self):
            self.pop_screen()

    counter = 0

    def reparse_wrapper(reparse):
        def _reparse(*args, **kwargs):
            nonlocal counter
            counter += 1
            return reparse(*args, **kwargs)

        return _reparse

    app = MyApp()
    app.stylesheet.reparse = reparse_wrapper(app.stylesheet.reparse)
    async with app.run_test() as pilot:
        await pilot.press("p")
        await pilot.press("o")
        await pilot.press("p")
        await pilot.press("o")
        await pilot.press("p")
        await pilot.press("o")
        await pilot.press("p")
        await pilot.press("o")
        assert counter == 1


async def test_screen_css_push_screen_instance():
    """Check that screen CSS is loaded and applied when pushing a screen instance."""

    class MyApp(BaseApp):
        def key_p(self):
            self.push_screen(ScreenWithCSS())

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_push_screen_instance_by_name():
    """Check that screen CSS is loaded and applied when pushing a screen name that points to a screen instance."""

    class MyApp(BaseApp):
        SCREENS = {"screenwithcss": ScreenWithCSS}

        def key_p(self):
            self.push_screen("screenwithcss")

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_push_screen_type_by_name():
    """Check that screen CSS is loaded and applied when pushing a screen name that points to a screen class."""

    class MyApp(BaseApp):
        SCREENS = {"screenwithcss": ScreenWithCSS}

        def key_p(self):
            self.push_screen("screenwithcss")

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_screen_instance():
    """Check that screen CSS is loaded and applied when switching to a screen instance."""

    class MyApp(SwitchBaseApp):
        def key_p(self):
            self.switch_screen(ScreenWithCSS())

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_screen_instance_by_name():
    """Check that screen CSS is loaded and applied when switching a screen name that points to a screen instance."""

    class MyApp(SwitchBaseApp):
        SCREENS = {"screenwithcss": ScreenWithCSS}

        def key_p(self):
            self.switch_screen("screenwithcss")

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_screen_type_by_name():
    """Check that screen CSS is loaded and applied when switching a screen name that points to a screen class."""

    class MyApp(SwitchBaseApp):
        SCREENS = {"screenwithcss": ScreenWithCSS}

        async def key_p(self):
            self.switch_screen("screenwithcss")

        def key_o(self):
            self.pop_screen()

    app = MyApp()
    async with app.run_test() as pilot:
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_mode_screen_instance():
    """Check that screen CSS is loaded and applied when switching to a mode with a screen instance."""

    class MyApp(BaseApp):
        MODES = {
            "base": BaseScreen,
            "mode": ScreenWithCSS,
        }

        def key_p(self):
            self.switch_mode("mode")

        def key_o(self):
            self.switch_mode("base")

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("o")
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_mode_screen_instance_by_name():
    """Check that screen CSS is loaded and applied when switching to a mode with a screen instance name."""

    class MyApp(BaseApp):
        SCREENS = {
            "screenwithcss": ScreenWithCSS,
        }

        MODES = {
            "base": BaseScreen,
            "mode": "screenwithcss",
        }

        def key_p(self):
            self.switch_mode("mode")

        def key_o(self):
            self.switch_mode("base")

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("o")
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)


async def test_screen_css_switch_mode_screen_type_by_name():
    """Check that screen CSS is loaded and applied when switching to a mode with a screen type name."""

    class MyApp(BaseApp):
        SCREENS = {
            "screenwithcss": ScreenWithCSS,
        }

        MODES = {
            "base": BaseScreen,
            "mode": "screenwithcss",
        }

        def key_p(self):
            self.switch_mode("mode")

        def key_o(self):
            self.switch_mode("base")

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("o")
        check_colors_before_screen_css(app)
        await pilot.press("p")
        check_colors_after_screen_css(app)
        await pilot.press("o")
        check_colors_after_screen_css(app)



import asyncio

from textual._context import active_message_pump
from textual.app import App
from textual.message_pump import MessagePump


class CallLaterApp(App[None]):
    def __init__(self) -> None:
        self.display_count = 0
        super().__init__()

    def post_display_hook(self) -> None:
        self.display_count += 1


async def test_call_later() -> None:
    """Check that call later makes a call."""
    app = CallLaterApp()
    called_event = asyncio.Event()

    async with app.run_test():
        app.call_later(called_event.set)
        await asyncio.wait_for(called_event.wait(), 1)


async def test_call_after_refresh() -> None:
    """Check that call later makes a call after a refresh."""
    app = CallLaterApp()

    display_count = -1

    called_event = asyncio.Event()

    callback_message_pump: MessagePump | None = None

    def callback() -> None:
        nonlocal display_count
        nonlocal callback_message_pump
        called_event.set()
        display_count = app.display_count
        callback_message_pump = active_message_pump.get()

    async with app.run_test():
        app.call_after_refresh(callback)
        await asyncio.wait_for(called_event.wait(), 1)
        app_display_count = app.display_count
        assert app_display_count == display_count
        # Should be app because the callback should be in the same context as app
        assert callback_message_pump is app



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import Button


class HorizontalWidthAutoApp(App[None]):
    CSS = """
    Horizontal {
        border: solid red;
        height: auto;
        width: auto;
    }
    """

    def compose(self) -> ComposeResult:
        with Horizontal(classes="auto"):
            yield Button("This is a very wide button")

        with Horizontal(classes="auto"):
            yield Button("This is a very wide button")
            yield Button("This is a very wide button")


if __name__ == "__main__":
    HorizontalWidthAutoApp().run()



from textual.widgets.text_area import Document

TEXT = """I must not fear.
Fear is the mind-killer."""


def test_insert_no_newlines():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), " really")
    assert document.lines == [
        "I really must not fear.",
        "Fear is the mind-killer.",
    ]


def test_insert_empty_string():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), "")
    assert document.lines == ["I must not fear.", "Fear is the mind-killer."]


def test_insert_invalid_column():
    document = Document(TEXT)
    document.replace_range((0, 999), (0, 999), " really")
    assert document.lines == ["I must not fear. really", "Fear is the mind-killer."]


def test_insert_invalid_row_and_column():
    document = Document(TEXT)
    document.replace_range((999, 0), (999, 0), " really")
    assert document.lines == ["I must not fear.", "Fear is the mind-killer.", " really"]


def test_insert_range_newline_file_start():
    document = Document(TEXT)
    document.replace_range((0, 0), (0, 0), "\n")
    assert document.lines == ["", "I must not fear.", "Fear is the mind-killer."]


def test_insert_newline_splits_line():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), "\n")
    assert document.lines == ["I", " must not fear.", "Fear is the mind-killer."]


def test_insert_newline_splits_line_selection():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 6), "\n")
    assert document.lines == ["I", " not fear.", "Fear is the mind-killer."]


def test_insert_multiple_lines_ends_with_newline():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), "Hello,\nworld!\n")
    assert document.lines == [
        "IHello,",
        "world!",
        " must not fear.",
        "Fear is the mind-killer.",
    ]


def test_insert_multiple_lines_ends_with_no_newline():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), "Hello,\nworld!")
    assert document.lines == [
        "IHello,",
        "world! must not fear.",
        "Fear is the mind-killer.",
    ]


def test_insert_multiple_lines_starts_with_newline():
    document = Document(TEXT)
    document.replace_range((0, 1), (0, 1), "\nHello,\nworld!\n")
    assert document.lines == [
        "I",
        "Hello,",
        "world!",
        " must not fear.",
        "Fear is the mind-killer.",
    ]


def test_insert_range_text_no_newlines():
    """Ensuring we can do a simple replacement of text."""
    document = Document(TEXT)
    document.replace_range((0, 2), (0, 6), "MUST")
    assert document.lines == [
        "I MUST not fear.",
        "Fear is the mind-killer.",
    ]


TEXT_NEWLINE_EOF = """\
I must not fear.
Fear is the mind-killer.
"""


def test_newline_eof():
    document = Document(TEXT_NEWLINE_EOF)
    assert document.lines == [
        "I must not fear.",
        "Fear is the mind-killer.",
        "",
    ]



from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Static


class UtilityContainersExample(App):
    CSS_PATH = "utility_containers.tcss"

    def compose(self) -> ComposeResult:
        with Horizontal():
            with Vertical(classes="column"):
                yield Static("One")
                yield Static("Two")
            with Vertical(classes="column"):
                yield Static("Three")
                yield Static("Four")


if __name__ == "__main__":
    app = UtilityContainersExample()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Static

TEXT = """
[b]Set your background[/b]
[@click=set_background('cyan')]Cyan[/]
[@click=set_background('magenta')]Magenta[/]
[@click=set_background('yellow')]Yellow[/]
"""


class ColorSwitcher(Static):
    def action_set_background(self, color: str) -> None:
        self.styles.background = color


class ActionsApp(App):
    CSS_PATH = "actions05.tcss"
    BINDINGS = [
        ("r", "set_background('red')", "Red"),
        ("g", "set_background('green')", "Green"),
        ("b", "set_background('blue')", "Blue"),
    ]

    def compose(self) -> ComposeResult:
        yield ColorSwitcher(TEXT)
        yield ColorSwitcher(TEXT)

    def action_set_background(self, color: str) -> None:
        self.screen.styles.background = color


if __name__ == "__main__":
    app = ActionsApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Static

TEXT1 = """\
Hello, [bold $text on $primary]World[/]!

[@click=app.notify('Hello, World!')]Click me[/]
"""

TEXT2 = """\
Markup will [bold]not[/bold] be displayed.

Tags will be left in the output.

"""


class ContentApp(App):
    CSS = """
    Screen {
        Static {
            height: 1fr;
        }
        #text1 { background: $primary-muted; }
        #text2 { background: $error-muted; }
    }
    """

    def compose(self) -> ComposeResult:
        yield Static(TEXT1, id="text1")
        yield Static(TEXT2, id="text2", markup=False)  # (1)!


if __name__ == "__main__":
    app = ContentApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import Static


class StaticText(Static):
    pass


class FRApp(App):
    CSS = """
    StaticText {
        height: 1fr;
        background: $boost;
        border: heavy white;
    }
    #foo {
        width: 10;
    }
    #bar {
        width: 1fr;
    }
    #baz {
        width: 8;
    }
    #header {
        height: 1fr
    }

    Horizontal {
        height: 2fr;
    }

    #footer {
        height: 4;
    }
    
    """

    def compose(self) -> ComposeResult:
        yield VerticalScroll(
            StaticText("HEADER", id="header"),
            Horizontal(
                StaticText("foo", id="foo"),
                StaticText("bar", id="bar"),
                StaticText("baz", id="baz"),
            ),
            StaticText("FOOTER", id="footer"),
        )


if __name__ == "__main__":
    app = FRApp()
    app.run()



from __future__ import annotations

import inspect

from rich.syntax import Syntax

from textual import work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import ScrollableContainer
from textual.screen import ModalScreen, Screen
from textual.widgets import Static


class CodeScreen(ModalScreen):
    DEFAULT_CSS = """
    CodeScreen {
        #code {
            border: heavy $accent;
            margin: 2 4;
            scrollbar-gutter: stable;
            Static {
                width: auto;
            }
        }
    }
    """
    BINDINGS = [("escape", "dismiss", "Dismiss code")]

    def __init__(self, title: str, code: str) -> None:
        super().__init__()
        self.code = code
        self.title = title

    def compose(self) -> ComposeResult:
        with ScrollableContainer(id="code"):
            yield Static(
                Syntax(
                    self.code, lexer="python", indent_guides=True, line_numbers=True
                ),
                expand=True,
            )

    def on_mount(self):
        code_widget = self.query_one("#code")
        code_widget.border_title = self.title
        code_widget.border_subtitle = "Escape to close"


class PageScreen(Screen):
    DEFAULT_CSS = """
    PageScreen {
        width: 100%;
        height: 1fr;
        overflow-y: auto;        
    }
    """
    BINDINGS = [
        Binding(
            "c",
            "show_code",
            "Code",
            tooltip="Show the code used to generate this screen",
        )
    ]

    @work(thread=True)
    def get_code(self, source_file: str) -> str | None:
        """Read code from disk, or return `None` on error."""
        try:
            with open(source_file, "rt", encoding="utf-8") as file_:
                return file_.read()
        except Exception:
            return None

    async def action_show_code(self):
        source_file = inspect.getsourcefile(self.__class__)
        if source_file is None:
            self.notify(
                "Could not get the code for this page",
                title="Show code",
                severity="error",
            )
            return

        code = await self.get_code(source_file).wait()
        if code is None:
            self.notify(
                "Could not get the code for this page",
                title="Show code",
                severity="error",
            )
        else:
            self.app.push_screen(CodeScreen("Code for this page", code))
